"""Power Supply Model classes."""

from . import pscreaders as _readers
from . import pscwriters as _writers
from . import pscontroller as _controller

from ..bsmp import constants as _const_psbsmp
from ..bsmp import entities as _etity_psbsmp


class _PSModel:
    """Abstract power supply model."""

    # class constant attributes to be overrriden in subclasses.
    _n = None  # PSModel name.
    _c = _const_psbsmp.ConstPSBSMP  # PSModel constantes.
    _e = _etity_psbsmp.EntitiesPS  # PSModel entities.

    # psmodel-specific conversions
    _bsmp_variables = dict()  # this will be filled in subclasses
    _pruc_properties = dict()  # this will be filled in subclasses

    # dictionary of epics PVs to common psmodel fields or variables.
    _e2v = {
        # Epics to direct BSMP variable
        'CycleEnbl-Mon': _c.V_SIGGEN_ENABLE,
        'CycleType-Sts': _c.V_SIGGEN_TYPE,
        'CycleNrCycles-RB': _c.V_SIGGEN_NUM_CYCLES,
        'CycleIndex-Mon': _c.V_SIGGEN_N,
        'CycleFreq-RB': _c.V_SIGGEN_FREQ,
        'CycleAmpl-RB': _c.V_SIGGEN_AMPLITUDE,
        'CycleOffset-RB': _c.V_SIGGEN_OFFSET,
        'CycleAuxParam-RB': _c.V_SIGGEN_AUX_PARAM,
        # Scope
        'ScopeSrcAddr-RB': _c.V_SCOPE_SRC_DATA,
        'ScopeFreq-RB': _c.V_SCOPE_FREQUENCY,
        'ScopeDuration-RB': _c.V_SCOPE_DURATION,
        # Trigger pulse diagnostics
        'NrCtrlCycBtwLastTrigs-Mon': _c.V_PERIOD_SYNC_PULSE,
        }
    _e2r = {
        # Epics to BSMP variable but with pre/post-processing
        'PwrState-Sts': (_readers.PwrState, _c.V_PS_STATUS),
        'OpMode-Sts': (_readers.OpMode, _c.V_PS_STATUS),
        'CtrlMode-Mon': (_readers.CtrlMode, _c.V_PS_STATUS),
        'CtrlLoop-Sts': (_readers.CtrlLoop, _c.V_PS_STATUS),
        'Version-Cte': (_readers.Version, _c.V_FIRMWARE_VERSION),
        }
    _e2c = {
        # Epics to PRUController property
        'PRUCtrlQueueSize-Mon': 'queue_length',
        }
    _e2o = {
        # Epics to reader object;
        'TimestampUpdate-Mon': _readers.TimestampUpdate,
        'Wfm-RB': _readers.WfmRBCurve,
        'WfmRef-Mon': _readers.WfmRefMonCurve,
        'Wfm-Mon': _readers.WfmMonCurve,
        'WfmIndex-Mon': _readers.WfmIndexCurve,
        'WfmUpdateAuto-Sts': _readers.WfmUpdateAutoSts,
        }
    _e2p = {
        # Epics to Power Supply parameter
        # --- PS ---
        'ParamPSName-Cte': _c.P_PS_NAME,  # 00
        'ParamPSModel-Cte': _c.P_PS_MODEL,  # 01
        'ParamNrModules-Cte': _c.P_PS_NR_PSMODELS,  # 02
        # --- Communication ---
        'ParamCommCmdInferface-Cte': _c.P_COMM_CMD_INTERFACE,  # 03
        'ParamCommRS485BaudRate-Cte': _c.P_COMM_RS485_BAUDRATE,  # 04
        'ParamCommRS485Addr-Cte': _c.P_COMM_RS485_ADDRESS,  # 05
        'ParamCommRS485TermRes-Cte': _c.P_COMM_RS485_TERMINATOR_RESISTOR,  # 06
        'ParamCommUDCNetAddr-Cte': _c.P_COMM_UDC_NETWORK_ADDRESS,  # 07
        'ParamCommEthIP-Cte': _c.P_COMM_ETHERNET_IP,  # 08
        'ParamCommEthSubnetMask-Cte': _c.P_COMM_ETHERNET_SUBNET_MASK,  # 09
        'ParamCommBuzVol-Cte': _c. P_COMM_BUZZER_VOLUME,  # 10
        # --- Control ---
        'ParamCtrlFreqCtrlISR-Cte': _c.P_CTRL_FREQ_CONTROL_ISR,  # 11
        'ParamCtrlFreqTimeSlicer-Cte': _c.P_CTRL_FREQ_TIME_SLICER,  # 12
        'ParamCtrlLoopState-Cte': _c.P_CTRL_CONTROL_LOOP_STATE,  # 13
        'ParamCtrlMaxRef-Cte': _c.P_CTRL_MAX_REF,  # 14
        'ParamCtrlMinRef-Cte': _c.P_CTRL_MIN_REF,  # 15
        'ParamCtrlMaxRefOpenLoop-Cte': _c.P_CTRL_MAX_REF_OPEN_LOOP,  # 16
        'ParamCtrlMinRefOpenLoop-Cte': _c.P_CTRL_MIN_REF_OPEN_LOOP,  # 17
        # --- PWM ---
        'ParamPWMFreq-Cte': _c.P_PWM_FREQ,  # 18
        'ParamPWMDeadTime-Cte': _c.P_PWM_DEAD_TIME,  # 19
        'ParamPWMMaxDuty-Cte': _c.P_PWM_MAX_DUTY,  # 20
        'ParamPWMMinDuty-Cte': _c.P_PWM_MIN_DUTY,  # 21
        'ParamPWMMaxDutyOpenLoop-Cte': _c.P_PWM_MAX_DUTY_OPEN_LOOP,  # 22
        'ParamPWMMinDutyOpenLoop-Cte': _c.P_PWM_MIN_DUTY_OPEN_LOOP,  # 23
        'ParamPWMLimDutyShare-Cte': _c.P_PWM_LIM_DUTY_SHARE,  # 24
        # ----- class HRADC -----
        'ParamHRADCNrBoards-Cte': _c.P_HRADC_NR_BOARDS,  # 25
        'ParamHRADCSpiClk-Cte': _c.P_HRADC_SPI_CLK,  # 26
        'ParamHRADCFreqSampling-Cte': _c.P_HRADC_FREQ_SAMPLING,  # 27
        'ParamHRADCEnableHeater-Cte': _c.P_HRADC_ENABLE_HEATER,  # 28
        'ParamHRADCEnableRails-Cte': _c.P_HRADC_ENABLE_RAILS_MON,  # 29
        'ParamHRADCTransducerOutput-Cte': _c.P_HRADC_TRANSDUCER_OUTPUT,  # 30
        'ParamHRADCTransducerGain-Cte': _c.P_HRADC_TRANSDUCER_GAIN,  # 31
        'ParamHRADCTransducerOffset-Cte': _c.P_HRADC_TRANSDUCER_OFFSET,  # 32
        # ----- class SigGen -----
        'ParamSigGenType-Cte': _c.P_SIGGEN_TYPE,  # 33
        'ParamSigGenNumCycles-Cte': _c.P_SIGGEN_NUM_CYCLES,  # 34
        'ParamSigGenFreq-Cte': _c.P_SIGGEN_FREQ,  # 35
        'ParamSigGenAmplitude-Cte': _c.P_SIGGEN_AMPLITUDE,  # 36
        'ParamSigGenOffset-Cte': _c.P_SIGGEN_OFFSET,  # 37
        'ParamSigGenAuxParam-Cte': _c.P_SIGGEN_AUX_PARAM,  # 38
        # ----- class WfmRef -----
        'ParamWfmRefId-Cte': _c.P_WFMREF_ID,  # 39
        'ParamWfmRefSyncMode-Cte': _c.P_WFMREF_SYNC_MODE,  # 40
        'ParamWfmRefFreq-Cte': _c.P_WFMREF_FREQ,  # 41
        'ParamWfmRefGain-Cte': _c.P_WFMREF_GAIN,  # 42
        'ParamWfmRefOffset-Cte': _c.P_WFMREF_OFFSET,  # 43
        # --- Analog Variables ---
        'ParamAnalogMax-Cte': _c.P_ANALOG_MAX,  # 44
        'ParamAnalogMin-Cte': _c.P_ANALOG_MIN,  # 45
        # --- Debounce Manager ---
        'ParamHardIntlkDebounceTime-Cte': _c.P_HARD_INTLK_DEBOUNCE_TIME,  # 46
        'ParamHardIntlkResetTime-Cte': _c.P_HARD_INTLK_RESET_TIME,  # 47
        'ParamSoftIntlkDebounceTime-Cte': _c.P_SOFT_INTLK_DEBOUNCE_TIME,  # 48
        'ParamSoftIntlkResetTime-Cte': _c.P_SOFT_INTLK_RESET_TIME,  # 49
        # ---- Scope -----
        'ParamScopeSamplingFreq-Cte': _c.P_SCOPE_SAMPLING_FREQUENCY,  # 50
        'ParamScopeDataSource-Cte': _c.P_SCOPE_DATA_SOURCE,  # 51
        }

    @property
    def name(self):
        """Model name."""
        return self._n

    @property
    def bsmp_constants(self):
        """PRU Controller parameters."""
        return self._c

    @property
    def entities(self):
        """PRU Controller parameters."""
        return self._e()

    def reader(self, device_id, epics_field, pru_controller):
        """Return reader."""
        # common reader
        e2f = self._readers_common
        reader = e2f(device_id, epics_field, pru_controller)
        if reader:
            return reader
        # especific reader (bsmp variable)
        e2f = self._readers_psmodel_specific_bsmp_variables
        reader = e2f(device_id, epics_field, pru_controller)
        if reader:
            return reader
        # specific reader (PRUController property)
        e2f = self._readers_psmodel_specific_pruc_properties
        reader = e2f(device_id, epics_field, pru_controller)
        return reader

    @staticmethod
    def writer(device_ids, epics_field, pru_controller, setpoints):
        """Return writer."""
        writer = _PSModel._writer_common_sp(
            device_ids, epics_field, pru_controller, setpoints)
        if writer:
            return writer

        writer = _PSModel._writer_common_cmd(
            device_ids, epics_field, pru_controller, setpoints)
        if writer:
            return writer

        writer = _PSModel._writer_cfgsiggen(
            device_ids, epics_field, pru_controller, setpoints)
        if writer:
            return writer

        writer = _PSModel._writer_wfm(
            device_ids, epics_field, pru_controller, setpoints)
        if writer:
            return writer

        return _writers.BSMPFunctionNull()

    def controller(self, readers, writers, pru_controller, devices):
        """Return controller."""
        return _controller.StandardPSController(
            readers, writers, pru_controller, devices)

    # --- private methods ---

    def _readers_common(self, device_id, epics_field, pru_controller):
        if epics_field in self._e2v:
            var_id = self._e2v[epics_field]
            return _readers.Variable(pru_controller, device_id, var_id)
        if epics_field in self._e2p:
            param_id = self._e2p[epics_field]
            return _readers.ConstParameter(pru_controller, device_id, param_id)
        if epics_field in self._e2r:
            reader, var_id = self._e2r[epics_field]
            return reader(_readers.Variable(pru_controller, device_id, var_id))
        if epics_field in self._e2c:
            attr = self._e2c[epics_field]
            return _readers.PRUCProperty(pru_controller, attr)
        if epics_field in self._e2o:
            reader = self._e2o[epics_field]
            return reader(pru_controller, device_id)
        return None

    def _readers_psmodel_specific_bsmp_variables(
            self, device_id, epics_field, pru_controller):
        # Specific fields
        if epics_field in self._bsmp_variables:
            var_id = self._bsmp_variables[epics_field]
            return _readers.Variable(pru_controller, device_id, var_id)
        return None

    def _readers_psmodel_specific_pruc_properties(
            self, _, epics_field, pru_controller):
        if epics_field in self._pruc_properties:
            attr = self._pruc_properties[epics_field]
            return _readers.PRUCProperty(pru_controller, attr)

    @staticmethod
    def _writer_common_sp(
            device_ids, epics_field, pru_controller, setpoints):
        _c = _const_psbsmp.ConstPSBSMP
        if epics_field == 'PwrState-Sel':
            return _writers.PSPwrState(device_ids, pru_controller, setpoints)
        elif epics_field == 'OpMode-Sel':
            bsmpfunc = _writers.BSMPFunction(
                device_ids, pru_controller, _c.F_SELECT_OP_MODE)
            return _writers.PSOpMode(
                device_ids, bsmpfunc, setpoints)
        elif epics_field == 'CtrlLoop-Sel':
            return _writers.CtrlLoop(device_ids, pru_controller, setpoints)
        elif epics_field == 'Current-SP':
            return _writers.Current(device_ids, pru_controller, setpoints)
        elif epics_field == 'ScopeFreq-SP':
            return _writers.ScopeFreq(device_ids, pru_controller, setpoints)
        elif epics_field == 'ScopeDuration-SP':
            return _writers.ScopeDuration(device_ids, pru_controller, setpoints)
        elif epics_field == 'ScopeSrcAddr-SP':
            return _writers.ScopeSrcAddr(device_ids, pru_controller, setpoints)
        return None

    @staticmethod
    def _writer_common_cmd(
            device_ids, epics_field, pru_controller, setpoints):
        _c = _const_psbsmp.ConstPSBSMP
        if epics_field == 'Reset-Cmd':
            return _writers.Command(
                device_ids, pru_controller, _c.F_RESET_INTERLOCKS, setpoints)
        if epics_field == 'SyncPulse-Cmd':
            return _writers.Command(
                device_ids, pru_controller, _c.F_SYNC_PULSE, setpoints)
        if epics_field == 'ParamUpdate-Cmd':
            return _writers.ParamUpdate(device_ids, pru_controller, setpoints)
        if epics_field == 'Abort-Cmd':
            return _writers.BSMPFunctionNull()
        return None

    @staticmethod
    def _writer_cfgsiggen(
            device_ids, epics_field, pru_controller, setpoints):
        p2i = {
            'CycleType-Sel': 0,
            'CycleNrCycles-SP': 1,
            'CycleFreq-SP': 2,
            'CycleAmpl-SP': 3,
            'CycleOffset-SP': 4,
            'CycleAuxParam-SP': 5,
            }
        _c = _const_psbsmp.ConstPSBSMP
        if epics_field in p2i:
            idx = p2i[epics_field]
            return _writers.CfgSiggen(
                device_ids, pru_controller, idx, setpoints)
        if epics_field == 'CycleDsbl-Cmd':
            return _writers.Command(
                device_ids, pru_controller, _c.F_DISABLE_SIGGEN, setpoints)
        return None

    @staticmethod
    def _writer_wfm(
            device_ids, epics_field, pru_controller, setpoints):
        if epics_field == 'Wfm-SP':
            return _writers.WfmSP(
                device_ids, pru_controller, setpoints)
        if epics_field == 'WfmUpdate-Cmd':
            return _writers.WfmUpdate(
                device_ids, pru_controller, setpoints)
        if epics_field == 'WfmUpdateAuto-Sel':
            return _writers.WfmUpdateAutoSel(
                device_ids, pru_controller, setpoints)
        if epics_field == 'WfmMonAcq-Sel':
            return _writers.WfmMonAcq(
                device_ids, pru_controller, setpoints)
        return None


# Standard PS that supply magnets
class PSModelFBP(_PSModel):
    """FBP model."""

    _n = 'FBP'
    _c = _const_psbsmp.ConstFBP
    _e = _etity_psbsmp.EntitiesFBP

    _bsmp_variables = {
        'IntlkSoft-Mon':  _c.V_PS_SOFT_INTERLOCKS,
        'IntlkHard-Mon':  _c.V_PS_HARD_INTERLOCKS,
        'WfmSyncPulseCount-Mon': _c.V_COUNTER_SYNC_PULSE,
        'Current-RB':  _c.V_PS_SETPOINT,
        'CurrentRef-Mon':  _c.V_PS_REFERENCE,
        'Current-Mon':  _c.V_I_LOAD,
        'LoadVoltage-Mon': _c.V_V_LOAD,
        'DCLinkVoltage-Mon': _c.V_V_DCLINK,
        'SwitchesTemperature-Mon': _c.V_TEMP_SWITCHES,
        'PWMDutyCycle-Mon': _c.V_DUTY_CYCLE,
        'Alarms-Mon': _c.V_PS_ALARMS,
    }

    _pruc_properties = {
        'SOFBMode-Sts': 'sofb_mode',
        'SOFBCurrent-RB': 'sofb_current_rb',
        'SOFBCurrentRef-Mon': 'sofb_current_refmon',
        'SOFBCurrent-Mon': 'sofb_current_mon',
    }

    def writer(self, device_ids, epics_field, pru_controller, setpoints):
        """Return writer."""
        if epics_field == 'SOFBCurrent-SP':
            return _writers.SOFBCurrent(
                device_ids, pru_controller, setpoints)
        if epics_field == 'SOFBMode-Sel':
            return _writers.SOFBMode(pru_controller, setpoints)
        if epics_field == 'SOFBUpdate-Cmd':
            return _writers.SOFBUpdate(pru_controller, setpoints)
        return super().writer(
            device_ids, epics_field, pru_controller, setpoints)


class PSModelFAC_DCDC(_PSModel):
    """FAC power supply model."""

    _n = 'FAC_DCDC'
    _c = _const_psbsmp.ConstFAC_DCDC
    _e = _etity_psbsmp.EntitiesFAC_DCDC

    _bsmp_variables = {
        'Current-RB': _c.V_PS_SETPOINT,
        'CurrentRef-Mon': _c.V_PS_REFERENCE,
        'WfmSyncPulseCount-Mon': _c.V_COUNTER_SYNC_PULSE,
        'IntlkSoft-Mon': _c.V_PS_SOFT_INTERLOCKS,
        'IntlkHard-Mon': _c.V_PS_HARD_INTERLOCKS,
        'Current-Mon': _c.V_I_LOAD_MEAN,
        'Current1-Mon': _c.V_I_LOAD1,
        'Current2-Mon': _c.V_I_LOAD2,
        'CapacitorBankVoltage-Mon': _c.V_V_CAPBANK,
        'PWMDutyCycle-Mon': _c.V_DUTY_CYCLE,
        'LeakCurrent-Mon': _c.V_I_LEAKAGE,
        'VoltageInputIIB-Mon': _c.V_V_INPUT_IIB,
        'CurrentInputIIB-Mon': _c.V_I_INPUT_IIB,
        'CurrentOutputIIB-Mon': _c.V_I_OUTPUT_IIB,
        'IGBT1TemperatureIIB-Mon': _c.V_TEMP_IGBT_1_IIB,
        'IGBT2TemperatureIIB-Mon': _c.V_TEMP_IGBT_2_IIB,
        'InductorsTemperatureIIB-Mon': _c.V_TEMP_INDUCTOR_IIB,
        'HeatSinkTemperatureIIB-Mon': _c.V_TEMP_HEATSINK_IIB,
        'IGBTDriverVoltageIIB-Mon': _c.V_V_DRIVER_IIB,
        'IGBT1DriverCurrentIIB-Mon': _c.V_I_DRIVER_1_IIB,
        'IGBT2DriverCurrentIIB-Mon': _c.V_I_DRIVER_2_IIB,
        'BoardTemperatureIIB-Mon': _c.V_TEMP_BOARD_IIB,
        'RelativeHumidityIIB-Mon': _c.V_RH_IIB,
        'IntlkIIB-Mon': _c.V_IIB_INTERLOCKS,
        'AlarmsIIB-Mon': _c.V_IIB_ALARMS,
        'Alarms-Mon': _c.V_PS_ALARMS,
    }


class PSModelFAC_2S_DCDC(_PSModel):
    """FAC_2S_DCDC power supply model."""

    _n = 'FAC_2S_DCDC'
    _c = _const_psbsmp.ConstFAC_2S_DCDC
    _e = _etity_psbsmp.EntitiesFAC_2S_DCDC

    _bsmp_variables = {
        'Current-RB': _c.V_PS_SETPOINT,
        'CurrentRef-Mon': _c.V_PS_REFERENCE,
        'WfmSyncPulseCount-Mon': _c.V_COUNTER_SYNC_PULSE,
        'IntlkSoft-Mon': _c.V_PS_SOFT_INTERLOCKS,
        'IntlkHard-Mon': _c.V_PS_HARD_INTERLOCKS,
        'Current-Mon': _c.V_I_LOAD_MEAN,
        'Current1-Mon': _c.V_I_LOAD1,
        'Current2-Mon': _c.V_I_LOAD2,
        'CapacitorBankVoltageMod1-Mon': _c.V_V_CAPBANK_1,
        'CapacitorBankVoltageMod2-Mon': _c.V_V_CAPBANK_2,
        'PWMDutyCycleMod1-Mon': _c.V_DUTY_CYCLE_1,
        'PWMDutyCycleMod2-Mon': _c.V_DUTY_CYCLE_2,
        'VoltageInputIIBMod1-Mon': _c.V_V_INPUT_IIB_1,
        'CurrentInputIIBMod1-Mon': _c.V_I_INPUT_IIB_1,
        'CurrentOutputIIBMod1-Mon': _c.V_I_OUTPUT_IIB_1,
        'IGBT1TemperatureIIBMod1-Mon': _c.V_TEMP_IGBTS_1_IIB_1,
        'IGBT2TemperatureIIBMod1-Mon': _c.V_TEMP_IGBTS_2_IIB_1,
        'InductorsTemperatureIIBMod1-Mon': _c.V_TEMP_INDUCTOR_IIB_1,
        'HeatSinkTemperatureIIBMod1-Mon': _c.V_TEMP_HEATSINK_IIB_1,
        'IGBTDriverVoltageIIBMod1-Mon': _c.V_V_DRIVER_IIB_1,
        'IGBT1DriverCurrentIIBMod1-Mon': _c.V_I_DRIVER_1_IIB_1,
        'IGBT2DriverCurrentIIBMod1-Mon': _c.V_I_DRIVER_2_IIB_1,
        'BoardTemperatureIIBMod1-Mon': _c.V_TEMP_BOARD_IIB_1,
        'RelativeHumidityIIBMod1-Mon': _c.V_RH_IIB_1,
        'IntlkIIBMod1-Mon': _c.V_IIB_INTERLOCKS_1,
        'AlarmsIIBMod1-Mon': _c.V_IIB_ALARMS_1,
        'VoltageInputIIBMod2-Mon': _c.V_V_INPUT_IIB_2,
        'CurrentInputIIBMod2-Mon': _c.V_I_INPUT_IIB_2,
        'CurrentOutputIIBMod2-Mon': _c.V_I_OUTPUT_IIB_2,
        'IGBT1TemperatureIIBMod2-Mon': _c.V_TEMP_IGBTS_1_IIB_2,
        'IGBT2TemperatureIIBMod2-Mon': _c.V_TEMP_IGBTS_2_IIB_2,
        'InductorsTemperatureIIBMod2-Mon': _c.V_TEMP_INDUCTOR_IIB_2,
        'HeatSinkTemperatureIIBMod2-Mon': _c.V_TEMP_HEATSINK_IIB_2,
        'IGBTDriverVoltageIIBMod2-Mon': _c.V_V_DRIVER_IIB_2,
        'IGBT1DriverCurrentIIBMod2-Mon': _c.V_I_DRIVER_1_IIB_2,
        'IGBT2DriverCurrentIIBMod2-Mon': _c.V_I_DRIVER_2_IIB_2,
        'BoardTemperatureIIBMod2-Mon': _c.V_TEMP_BOARD_IIB_2,
        'RelativeHumidityIIBMod2-Mon': _c.V_RH_IIB_2,
        'IntlkIIBMod2-Mon': _c.V_IIB_INTERLOCKS_2,
        'AlarmsIIBMod2-Mon': _c.V_IIB_ALARMS_2,
        'Alarms-Mon': _c.V_PS_ALARMS,
        }


class PSModelFAC_2P4S_DCDC(PSModelFAC_DCDC):
    """FAC_2P4S_DCDC power supply model (BO Dipoles)."""

    _n = 'FAC_2P4S_DCDC'
    _c = _const_psbsmp.ConstFAC_2P4S_DCDC
    _e = _etity_psbsmp.EntitiesFAC_2P4S_DCDC

    _bsmp_variables = {
        'Current-RB': _c.V_PS_SETPOINT,
        'CurrentRef-Mon': _c.V_PS_REFERENCE,
        'WfmSyncPulseCount-Mon': _c.V_COUNTER_SYNC_PULSE,
        'IntlkSoft-Mon': _c.V_PS_SOFT_INTERLOCKS,
        'IntlkHard-Mon': _c.V_PS_HARD_INTERLOCKS,
        'Current-Mon': _c.V_I_LOAD_MEAN,
        'Current1-Mon': _c.V_I_LOAD1,
        'Current2-Mon': _c.V_I_LOAD2,
        'Arm1Current-Mon': _c.V_I_ARM_1,
        'Arm2Current-Mon': _c.V_I_ARM_2,
        'CapacitorBankVoltageMod1-Mon': _c.V_V_CAPBANK_1,
        'CapacitorBankVoltageMod2-Mon': _c.V_V_CAPBANK_2,
        'CapacitorBankVoltageMod3-Mon': _c.V_V_CAPBANK_3,
        'CapacitorBankVoltageMod4-Mon': _c.V_V_CAPBANK_4,
        'CapacitorBankVoltageMod5-Mon': _c.V_V_CAPBANK_5,
        'CapacitorBankVoltageMod6-Mon': _c.V_V_CAPBANK_6,
        'CapacitorBankVoltageMod7-Mon': _c.V_V_CAPBANK_7,
        'CapacitorBankVoltageMod8-Mon': _c.V_V_CAPBANK_8,
        'PWMDutyCycleMod1-Mon': _c.V_DUTY_CYCLE_1,
        'PWMDutyCycleMod2-Mon': _c.V_DUTY_CYCLE_2,
        'PWMDutyCycleMod3-Mon': _c.V_DUTY_CYCLE_3,
        'PWMDutyCycleMod4-Mon': _c.V_DUTY_CYCLE_4,
        'PWMDutyCycleMod5-Mon': _c.V_DUTY_CYCLE_5,
        'PWMDutyCycleMod6-Mon': _c.V_DUTY_CYCLE_6,
        'PWMDutyCycleMod7-Mon': _c.V_DUTY_CYCLE_7,
        'PWMDutyCycleMod8-Mon': _c.V_DUTY_CYCLE_8,
        'VoltageInputIIBModA-Mon': _c.V_V_INPUT_IIB_A,
        'CurrentInputIIBModA-Mon': _c.V_I_INPUT_IIB_A,
        'CurrentOutputIIBModA-Mon': _c.V_I_OUTPUT_IIB_A,
        'IGBT1TemperatureIIBModA-Mon': _c.V_TEMP_IGBTS_1_IIB_A,
        'IGBT2TemperatureIIBModA-Mon': _c.V_TEMP_IGBTS_2_IIB_A,
        'InductorTemperatureIIBModA-Mon': _c.V_TEMP_INDUCTOR_IIB_A,
        'HeatSinkTemperatureIIBModA-Mon': _c.V_TEMP_HEATSINK_IIB_A,
        'IGBTDriverVoltageIIBModA-Mon': _c.V_V_DRIVER_IIB_A,
        'IGBT1DriverCurrentIIBModA-Mon': _c.V_I_DRIVER_1_IIB_A,
        'IGBT2DriverCurrentIIBModA-Mon': _c.V_I_DRIVER_2_IIB_A,
        'BoardTemperatureIIBModA-Mon': _c.V_TEMP_BOARD_IIB_A,
        'RelativeHumidityIIBModA-Mon': _c.V_RH_IIB_A,
        'IntlkIIBModA-Mon': _c.V_IIB_INTERLOCKS_A,
        'AlarmsIIBModA-Mon': _c.V_IIB_ALARMS_A,
        'VoltageInputIIBModB-Mon': _c.V_V_INPUT_IIB_B,
        'CurrentInputIIBModB-Mon': _c.V_I_INPUT_IIB_B,
        'CurrentOutputIIBModB-Mon': _c.V_I_OUTPUT_IIB_B,
        'IGBT1TemperatureIIBModB-Mon': _c.V_TEMP_IGBTS_1_IIB_B,
        'IGBT2TemperatureIIBModB-Mon': _c.V_TEMP_IGBTS_2_IIB_B,
        'InductorTemperatureIIBModB-Mon': _c.V_TEMP_INDUCTOR_IIB_B,
        'HeatSinkTemperatureIIBModB-Mon': _c.V_TEMP_HEATSINK_IIB_B,
        'IGBTDriverVoltageIIBModB-Mon': _c.V_V_DRIVER_IIB_B,
        'IGBT1DriverCurrentIIBModB-Mon': _c.V_I_DRIVER_1_IIB_B,
        'IGBT2DriverCurrentIIBModB-Mon': _c.V_I_DRIVER_2_IIB_B,
        'BoardTemperatureIIBModB-Mon': _c.V_TEMP_BOARD_IIB_B,
        'RelativeHumidityIIBModB-Mon': _c.V_RH_IIB_B,
        'IntlkIIBModB-Mon': _c.V_IIB_INTERLOCKS_B,
        'AlarmsIIBModB-Mon': _c.V_IIB_ALARMS_B,
        'Alarms-Mon': _c.V_PS_ALARMS,
        }


class PSModelFAP(_PSModel):
    """FAP power supply model."""

    _n = 'FAP'
    _c = _const_psbsmp.ConstFAP
    _e = _etity_psbsmp.EntitiesFAP

    _bsmp_variables = {
        'Current-RB': _c.V_PS_SETPOINT,
        'CurrentRef-Mon': _c.V_PS_REFERENCE,
        'WfmSyncPulseCount-Mon': _c.V_COUNTER_SYNC_PULSE,
        'IntlkSoft-Mon': _c.V_PS_SOFT_INTERLOCKS,
        'IntlkHard-Mon': _c.V_PS_HARD_INTERLOCKS,
        'Current-Mon': _c.V_I_LOAD_MEAN,
        'Current1-Mon': _c.V_I_LOAD1,
        'Current2-Mon': _c.V_I_LOAD2,
        'DCLinkVoltage-Mon': _c.V_V_DCLINK,
        'IGBT1Current-Mon': _c.V_I_IGBT_1,
        'IGBT2Current-Mon': _c.V_I_IGBT_2,
        'IGBT1PWMDutyCycle-Mon': _c.V_DUTY_CYCLE_1,
        'IGBT2PWMDutyCycle-Mon': _c.V_DUTY_CYCLE_2,
        'PWMDutyDiff-Mon': _c.V_DUTY_DIFF,
        'VoltageInputIIB-Mon': _c.V_V_INPUT_IIB,
        'VoltageOutputIIB-Mon': _c.V_V_OUTPUT_IIB,
        'IGBT1CurrentIIB-Mon': _c.V_I_IGBT_1_IIB,
        'IGBT2CurrentIIB-Mon': _c.V_I_IGBT_2_IIB,
        'IGBT1TemperatureIIB-Mon': _c.V_TEMP_IGBT_1_IIB,
        'IGBT2TemperatureIIB-Mon': _c.V_TEMP_IGBT_2_IIB,
        'IGBTDriverVoltageIIB-Mon': _c.V_V_DRIVER_IIB,
        'IGBT1DriverCurrentIIB-Mon': _c.V_I_DRIVER_1_IIB,
        'IGBT2DriverCurrentIIB-Mon': _c.V_I_DRIVER_2_IIB,
        'InductorTemperatureIIB-Mon': _c.V_TEMP_INDUCTOR_IIB,
        'HeatSinkTemperatureIIB-Mon': _c.V_TEMP_HEATSINK_IIB,
        'LeakCurrentIIB-Mon': _c.V_I_LEAKAGE_IIB,
        'BoardTemperatureIIB-Mon': _c.V_TEMP_BOARD_IIB,
        'RelativeHumidityIIB-Mon': _c.V_RH_IIB,
        'IntlkIIB-Mon': _c.V_IIB_INTERLOCKS,
        'AlarmsIIB-Mon': _c.V_IIB_ALARMS,
        'Alarms-Mon': _c.V_PS_ALARMS,
        }


class PSModelFAP_4P(_PSModel):
    """FAP_4P power supply model."""

    _n = 'FAP_4P'
    _c = _const_psbsmp.ConstFAP_4P
    _e = _etity_psbsmp.EntitiesFAP_4P

    _bsmp_variables = {
        'Current-RB': _c.V_PS_SETPOINT,
        'CurrentRef-Mon': _c.V_PS_REFERENCE,
        'WfmSyncPulseCount-Mon': _c.V_COUNTER_SYNC_PULSE,
        'IntlkSoft-Mon': _c.V_PS_SOFT_INTERLOCKS,
        'IntlkHard-Mon': _c.V_PS_HARD_INTERLOCKS,
        'Current1-Mon': _c.V_I_LOAD1,
        'Current2-Mon': _c.V_I_LOAD2,
        'Current-Mon': _c.V_I_LOAD_MEAN,
        'LoadVoltage-Mon': _c.V_V_LOAD,
        'IGBT1CurrentMod1-Mon': _c.V_I_IGBT_1_1,
        'IGBT2CurrentMod1-Mon': _c.V_I_IGBT_2_1,
        'IGBT1CurrentMod2-Mon': _c.V_I_IGBT_1_2,
        'IGBT2CurrentMod2-Mon': _c.V_I_IGBT_2_2,
        'IGBT1CurrentMod3-Mon': _c.V_I_IGBT_1_3,
        'IGBT2CurrentMod3-Mon': _c.V_I_IGBT_2_3,
        'IGBT1CurrentMod4-Mon': _c.V_I_IGBT_1_4,
        'IGBT2CurrentMod4-Mon': _c.V_I_IGBT_2_4,
        'DCLinkVoltageMod1-Mon': _c.V_V_DCLINK_1,
        'DCLinkVoltageMod2-Mon': _c.V_V_DCLINK_2,
        'DCLinkVoltageMod3-Mon': _c.V_V_DCLINK_3,
        'DCLinkVoltageMod4-Mon': _c.V_V_DCLINK_4,
        'PWMDutyCycle-Mon': _c.V_DUTY_MEAN,
        'IGBT1PWMDutyCycleMod1-Mon': _c.V_DUTY_CYCLE_1_1,
        'IGBT2PWMDutyCycleMod1-Mon': _c.V_DUTY_CYCLE_2_1,
        'IGBT1PWMDutyCycleMod2-Mon': _c.V_DUTY_CYCLE_1_2,
        'IGBT2PWMDutyCycleMod2-Mon': _c.V_DUTY_CYCLE_2_2,
        'IGBT1PWMDutyCycleMod3-Mon': _c.V_DUTY_CYCLE_1_3,
        'IGBT2PWMDutyCycleMod3-Mon': _c.V_DUTY_CYCLE_2_3,
        'IGBT1PWMDutyCycleMod4-Mon': _c.V_DUTY_CYCLE_1_4,
        'IGBT2PWMDutyCycleMod4-Mon': _c.V_DUTY_CYCLE_2_4,
        'LeakCurrent-Mon': _c.V_I_LEAKAGE,
        'VoltageInputIIBMod1-Mon': _c.V_V_INPUT_IIB_1,
        'VoltageOutputIIBMod1-Mon': _c.V_V_OUTPUT_IIB_1,
        'IGBT1CurrentIIBMod1-Mon': _c.V_I_IGBT_1_IIB_1,
        'IGBT2CurrentIIBMod1-Mon': _c.V_I_IGBT_2_IIB_1,
        'IGBT1TemperatureIIBMod1-Mon': _c.V_TEMP_IGBT_1_IIB_1,
        'IGBT2TemperatureIIBMod1-Mon': _c.V_TEMP_IGBT_2_IIB_1,
        'IGBTDriverVoltageIIBMod1-Mon': _c.V_V_DRIVER_IIB_1,
        'IGBT1DriverCurrentIIBMod1-Mon': _c.V_I_DRIVER_1_IIB_1,
        'IGBT2DriverCurrentIIBMod1-Mon': _c.V_I_DRIVER_2_IIB_1,
        'InductorTemperatureIIBMod1-Mon': _c.V_TEMP_INDUCTOR_IIB_1,
        'HeatSinkTemperatureIIBMod1-Mon': _c.V_TEMP_HEATSINK_IIB_1,
        'TemperatureIIBMod1-Mon': _c.V_TEMP_BOARD_IIB_1,
        'RelativeHumidityIIBMod1-Mon': _c.V_RH_IIB_1,
        'IntlkIIBMod1-Mon': _c.V_IIB_INTERLOCKS_1,
        'AlarmsIIBMod1-Mon': _c.V_IIB_ALARMS_1,
        'VoltageInputIIBMod2-Mon': _c.V_V_INPUT_IIB_2,
        'VoltageOutputIIBMod2-Mon': _c.V_V_OUTPUT_IIB_2,
        'IGBT1CurrentIIBMod2-Mon': _c.V_I_IGBT_1_IIB_2,
        'IGBT2CurrentIIBMod2-Mon': _c.V_I_IGBT_2_IIB_2,
        'IGBT1TemperatureIIBMod2-Mon': _c.V_TEMP_IGBT_1_IIB_2,
        'IGBT2TemperatureIIBMod2-Mon': _c.V_TEMP_IGBT_2_IIB_2,
        'IGBTDriverVoltageIIBMod2-Mon': _c.V_V_DRIVER_IIB_2,
        'IGBT1DriverCurrentIIBMod2-Mon': _c.V_I_DRIVER_1_IIB_2,
        'IGBT2DriverCurrentIIBMod2-Mon': _c.V_I_DRIVER_2_IIB_2,
        'InductorTemperatureIIBMod2-Mon': _c.V_TEMP_INDUCTOR_IIB_2,
        'HeatSinkTemperatureIIBMod2-Mon': _c.V_TEMP_HEATSINK_IIB_2,
        'TemperatureIIBMod2-Mon': _c.V_TEMP_BOARD_IIB_2,
        'RelativeHumidityIIBMod2-Mon': _c.V_RH_IIB_2,
        'IntlkIIBMod2-Mon': _c.V_IIB_INTERLOCKS_2,
        'AlarmsIIBMod2-Mon': _c.V_IIB_ALARMS_2,
        'VoltageInputIIBMod3-Mon': _c.V_V_INPUT_IIB_3,
        'VoltageOutputIIBMod3-Mon': _c.V_V_OUTPUT_IIB_3,
        'IGBT1CurrentIIBMod3-Mon': _c.V_I_IGBT_1_IIB_3,
        'IGBT2CurrentIIBMod3-Mon': _c.V_I_IGBT_2_IIB_3,
        'IGBT1TemperatureIIBMod3-Mon': _c.V_TEMP_IGBT_1_IIB_3,
        'IGBT2TemperatureIIBMod3-Mon': _c.V_TEMP_IGBT_2_IIB_3,
        'IGBTDriverVoltageIIBMod3-Mon': _c.V_V_DRIVER_IIB_3,
        'IGBT1DriverCurrentIIBMod3-Mon': _c.V_I_DRIVER_1_IIB_3,
        'IGBT2DriverCurrentIIBMod3-Mon': _c.V_I_DRIVER_2_IIB_3,
        'InductorTemperatureIIBMod3-Mon': _c.V_TEMP_INDUCTOR_IIB_3,
        'HeatSinkTemperatureIIBMod3-Mon': _c.V_TEMP_HEATSINK_IIB_3,
        'TemperatureIIBMod3-Mon': _c.V_TEMP_BOARD_IIB_3,
        'RelativeHumidityIIBMod3-Mon': _c.V_RH_IIB_3,
        'IntlkIIBMod3-Mon': _c.V_IIB_INTERLOCKS_3,
        'AlarmsIIBMod3-Mon': _c.V_IIB_ALARMS_3,
        'VoltageInputIIBMod4-Mon': _c.V_V_INPUT_IIB_4,
        'VoltageOutputIIBMod4-Mon': _c.V_V_OUTPUT_IIB_4,
        'IGBT1CurrentIIBMod4-Mon': _c.V_I_IGBT_1_IIB_4,
        'IGBT2CurrentIIBMod4-Mon': _c.V_I_IGBT_2_IIB_4,
        'IGBT1TemperatureIIBMod4-Mon': _c.V_TEMP_IGBT_1_IIB_4,
        'IGBT2TemperatureIIBMod4-Mon': _c.V_TEMP_IGBT_2_IIB_4,
        'IGBTDriverVoltageIIBMod4-Mon': _c.V_V_DRIVER_IIB_4,
        'IGBT1DriverCurrentIIBMod4-Mon': _c.V_I_DRIVER_1_IIB_4,
        'IGBT2DriverCurrentIIBMod4-Mon': _c.V_I_DRIVER_2_IIB_4,
        'InductorTemperatureIIBMod4-Mon': _c.V_TEMP_INDUCTOR_IIB_4,
        'HeatSinkTemperatureIIBMod4-Mon': _c.V_TEMP_HEATSINK_IIB_4,
        'TemperatureIIBMod4-Mon': _c.V_TEMP_BOARD_IIB_4,
        'RelativeHumidityIIBMod4-Mon': _c.V_RH_IIB_4,
        'IntlkIIBMod4-Mon': _c.V_IIB_INTERLOCKS_4,
        'AlarmsIIBMod4-Mon': _c.V_IIB_ALARMS_4,
        'Alarms-Mon': _c.V_PS_ALARMS,
        }


class PSModelFAP_2P2S(_PSModel):
    """FAP_2P2S power supply model."""

    _n = 'FAP_2P2S'
    _c = _const_psbsmp.ConstFAP_2P2S
    _e = _etity_psbsmp.EntitiesFAP_2P2S

    _bsmp_variables = {
        'Current-RB':  _c.V_PS_SETPOINT,
        'CurrentRef-Mon':  _c.V_PS_REFERENCE,
        'WfmSyncPulseCount-Mon': _c.V_COUNTER_SYNC_PULSE,
        'IntlkSoft-Mon':  _c.V_PS_SOFT_INTERLOCKS,
        'IntlkHard-Mon':  _c.V_PS_HARD_INTERLOCKS,
        'Current-Mon':  _c.V_I_LOAD_MEAN,
        'Current1-Mon': _c.V_I_LOAD1,
        'Current2-Mon': _c.V_I_LOAD2,
        'Arm1Current-Mon': _c.V_I_ARM_1,
        'Arm2Current-Mon': _c.V_I_ARM_2,
        'IGBT1CurrentMod1-Mon': _c.V_I_IGBT_1_1,
        'IGBT2CurrentMod1-Mon': _c.V_I_IGBT_2_1,
        'IGBT1CurrentMod2-Mon': _c.V_I_IGBT_1_2,
        'IGBT2CurrentMod2-Mon': _c.V_I_IGBT_2_2,
        'IGBT1CurrentMod3-Mon': _c.V_I_IGBT_1_3,
        'IGBT2CurrentMod3-Mon': _c.V_I_IGBT_2_3,
        'IGBT1CurrentMod4-Mon': _c.V_I_IGBT_1_4,
        'IGBT2CurrentMod4-Mon': _c.V_I_IGBT_2_4,
        'CurrentMod1-Mon': _c.V_I_MOD_1,
        'CurrentMod2-Mon': _c.V_I_MOD_2,
        'CurrentMod3-Mon': _c.V_I_MOD_3,
        'CurrentMod4-Mon': _c.V_I_MOD_4,
        'DCLinkVoltageMod1-Mon': _c.V_V_DCLINK_1,
        'DCLinkVoltageMod2-Mon': _c.V_V_DCLINK_2,
        'DCLinkVoltageMod3-Mon': _c.V_V_DCLINK_3,
        'DCLinkVoltageMod4-Mon': _c.V_V_DCLINK_4,
        'PWMDutyCycle-Mon': _c.V_DUTY_MEAN,
        'PWMDutyCycleArmsDiff-Mon': _c.V_DUTY_ARMS_DIFF,
        'IGBT1PWMDutyCycleMod1-Mon': _c.V_DUTY_CYCLE_1_1,
        'IGBT2PWMDutyCycleMod1-Mon': _c.V_DUTY_CYCLE_2_1,
        'IGBT1PWMDutyCycleMod2-Mon': _c.V_DUTY_CYCLE_1_2,
        'IGBT2PWMDutyCycleMod2-Mon': _c.V_DUTY_CYCLE_2_2,
        'IGBT1PWMDutyCycleMod3-Mon': _c.V_DUTY_CYCLE_1_3,
        'IGBT2PWMDutyCycleMod3-Mon': _c.V_DUTY_CYCLE_2_3,
        'IGBT1PWMDutyCycleMod4-Mon': _c.V_DUTY_CYCLE_1_4,
        'IGBT2PWMDutyCycleMod4-Mon': _c.V_DUTY_CYCLE_2_4,
        'LeakCurrent-Mon': _c.V_I_LEAKAGE,
        'VoltageInputIIBMod1-Mon': _c.V_V_INPUT_IIB_1,
        'VoltageOutputIIBMod1-Mon': _c.V_V_OUTPUT_IIB_1,
        'IGBT1CurrentIIBMod1-Mon': _c.V_I_IGBT_1_IIB_1,
        'IGBT2CurrentIIBMod1-Mon': _c.V_I_IGBT_2_IIB_1,
        'IGBT1TemperatureIIBMod1-Mon': _c.V_TEMP_IGBT_1_IIB_1,
        'IGBT2TemperatureIIBMod1-Mon': _c.V_TEMP_IGBT_2_IIB_1,
        'IGBTDriverVoltageIIBMod1-Mon': _c.V_V_DRIVER_IIB_1,
        'IGBT1DriverCurrentIIBMod1-Mon': _c.V_I_DRIVER_1_IIB_1,
        'IGBT2DriverCurrentIIBMod1-Mon': _c.V_I_DRIVER_2_IIB_1,
        'InductorTemperatureIIBMod1-Mon': _c.V_TEMP_INDUCTOR_IIB_1,
        'HeatSinkTemperatureIIBMod1-Mon': _c.V_TEMP_HEATSINK_IIB_1,
        'TemperatureIIBMod1-Mon': _c.V_TEMP_BOARD_IIB_1,
        'RelativeHumidityIIBMod1-Mon': _c.V_RH_IIB_1,
        'IntlkIIBMod1-Mon': _c.V_IIB_INTERLOCKS_1,
        'AlarmsIIBMod1-Mon': _c.V_IIB_ALARMS_1,
        'VoltageInputIIBMod2-Mon': _c.V_V_INPUT_IIB_2,
        'VoltageOutputIIBMod2-Mon': _c.V_V_OUTPUT_IIB_2,
        'IGBT1CurrentIIBMod2-Mon': _c.V_I_IGBT_1_IIB_2,
        'IGBT2CurrentIIBMod2-Mon': _c.V_I_IGBT_2_IIB_2,
        'IGBT1TemperatureIIBMod2-Mon': _c.V_TEMP_IGBT_1_IIB_2,
        'IGBT2TemperatureIIBMod2-Mon': _c.V_TEMP_IGBT_2_IIB_2,
        'IGBTDriverVoltageIIBMod2-Mon': _c.V_V_DRIVER_IIB_2,
        'IGBT1DriverCurrentIIBMod2-Mon': _c.V_I_DRIVER_1_IIB_2,
        'IGBT2DriverCurrentIIBMod2-Mon': _c.V_I_DRIVER_2_IIB_2,
        'InductorTemperatureIIBMod2-Mon': _c.V_TEMP_INDUCTOR_IIB_2,
        'HeatSinkTemperatureIIBMod2-Mon': _c.V_TEMP_HEATSINK_IIB_2,
        'TemperatureIIBMod2-Mon': _c.V_TEMP_BOARD_IIB_2,
        'RelativeHumidityIIBMod2-Mon': _c.V_RH_IIB_2,
        'IntlkIIBMod2-Mon': _c.V_IIB_INTERLOCKS_2,
        'AlarmsIIBMod2-Mon': _c.V_IIB_ALARMS_2,
        'VoltageInputIIBMod3-Mon': _c.V_V_INPUT_IIB_3,
        'VoltageOutputIIBMod3-Mon': _c.V_V_OUTPUT_IIB_3,
        'IGBT1CurrentIIBMod3-Mon': _c.V_I_IGBT_1_IIB_3,
        'IGBT2CurrentIIBMod3-Mon': _c.V_I_IGBT_2_IIB_3,
        'IGBT1TemperatureIIBMod3-Mon': _c.V_TEMP_IGBT_1_IIB_3,
        'IGBT2TemperatureIIBMod3-Mon': _c.V_TEMP_IGBT_2_IIB_3,
        'IGBTDriverVoltageIIBMod3-Mon': _c.V_V_DRIVER_IIB_3,
        'IGBT1DriverCurrentIIBMod3-Mon': _c.V_I_DRIVER_1_IIB_3,
        'IGBT2DriverCurrentIIBMod3-Mon': _c.V_I_DRIVER_2_IIB_3,
        'InductorTemperatureIIBMod3-Mon': _c.V_TEMP_INDUCTOR_IIB_3,
        'HeatSinkTemperatureIIBMod3-Mon': _c.V_TEMP_HEATSINK_IIB_3,
        'TemperatureIIBMod3-Mon': _c.V_TEMP_BOARD_IIB_3,
        'RelativeHumidityIIBMod3-Mon': _c.V_RH_IIB_3,
        'IntlkIIBMod3-Mon': _c.V_IIB_INTERLOCKS_3,
        'AlarmsIIBMod3-Mon': _c.V_IIB_ALARMS_3,
        'VoltageInputIIBMod4-Mon': _c.V_V_INPUT_IIB_4,
        'VoltageOutputIIBMod4-Mon': _c.V_V_OUTPUT_IIB_4,
        'IGBT1CurrentIIBMod4-Mon': _c.V_I_IGBT_1_IIB_4,
        'IGBT2CurrentIIBMod4-Mon': _c.V_I_IGBT_2_IIB_4,
        'IGBT1TemperatureIIBMod4-Mon': _c.V_TEMP_IGBT_1_IIB_4,
        'IGBT2TemperatureIIBMod4-Mon': _c.V_TEMP_IGBT_2_IIB_4,
        'IGBTDriverVoltageIIBMod4-Mon': _c.V_V_DRIVER_IIB_4,
        'IGBT1DriverCurrentIIBMod4-Mon': _c.V_I_DRIVER_1_IIB_4,
        'IGBT2DriverCurrentIIBMod4-Mon': _c.V_I_DRIVER_2_IIB_4,
        'InductorTemperatureIIBMod4-Mon': _c.V_TEMP_INDUCTOR_IIB_4,
        'HeatSinkTemperatureIIBMod4-Mon': _c.V_TEMP_HEATSINK_IIB_4,
        'TemperatureIIBMod4-Mon': _c.V_TEMP_BOARD_IIB_4,
        'RelativeHumidityIIBMod4-Mon': _c.V_RH_IIB_4,
        'IntlkIIBMod4-Mon': _c.V_IIB_INTERLOCKS_4,
        'AlarmsIIBMod4-Mon': _c.V_IIB_ALARMS_4,
        'Alarms-Mon': _c.V_PS_ALARMS,
        }


# --- ACDC ---


class PSModelFBP_DCLink(_PSModel):
    """FBP_DCLink model."""

    _n = 'FBP_DCLink'
    _c = _const_psbsmp.ConstFBP_DCLink
    _e = _etity_psbsmp.EntitiesFBP_DCLink

    _bsmp_variables = {
        'IntlkSoft-Mon': _c.V_PS_SOFT_INTERLOCKS,
        'IntlkHard-Mon': _c.V_PS_HARD_INTERLOCKS,
        'ModulesStatus-Mon': _c.V_MODULES_STATUS,
        'Voltage-RB': _c.V_PS_SETPOINT,
        'VoltageRef-Mon': _c.V_PS_REFERENCE,
        'Voltage-Mon': _c.V_V_OUT,
        'Voltage1-Mon': _c.V_V_OUT_1,
        'Voltage2-Mon': _c.V_V_OUT_2,
        'Voltage3-Mon': _c.V_V_OUT_3,
        'VoltageDig-Mon': _c.V_DIG_POT_TAP,
        }

    def writer(self, device_ids, epics_field, pru_controller, setpoints):
        """Return writer."""
        _c = PSModelFBP_DCLink._c
        if epics_field == 'PwrState-Sel':
            return _writers.PSPwrStateFBP_DCLink(
                device_ids, pru_controller, setpoints)
        elif epics_field == 'CtrlLoop-Sel':
            return _writers.CtrlLoop(device_ids, pru_controller, setpoints)
        elif epics_field == 'OpMode-Sel':
            return _writers.PSOpMode(
                device_ids,
                _writers.BSMPFunction(
                    device_ids, pru_controller, _c.F_SELECT_OP_MODE),
                setpoints)
        elif epics_field == 'Voltage-SP':
            return _writers.BSMPFunction(
                device_ids, pru_controller, _c.F_SET_SLOWREF, setpoints)
        elif epics_field == 'Reset-Cmd':
            return _writers.Command(
                device_ids, pru_controller, _c.F_RESET_INTERLOCKS, setpoints)
        elif epics_field == 'Abort-Cmd':
            return _writers.BSMPFunctionNull()
        else:
            return _writers.BSMPFunctionNull()

    def controller(self, readers, writers, pru_controller, devices):
        """Return controller."""
        return _controller.PSController(
            readers, writers, pru_controller, devices)


class PSModelFAC_2S_ACDC(_PSModel):
    """FAC_2S_ACDC model."""

    _n = 'FAC_2S_ACDC'
    _c = _const_psbsmp.ConstFAC_2S_ACDC
    _e = _etity_psbsmp.EntitiesFAC_2S_ACDC

    _bsmp_variables = {
        'CapacitorBankVoltage-RB': _c.V_PS_SETPOINT,
        'CapacitorBankVoltageRef-Mon': _c.V_PS_REFERENCE,
        'IntlkSoft-Mon': _c.V_PS_SOFT_INTERLOCKS,
        'IntlkHard-Mon': _c.V_PS_HARD_INTERLOCKS,
        'CapacitorBankVoltage-Mon': _c.V_V_CAPBANK,
        'RectifierCurrent-Mon': _c.V_I_OUT_RECTIFIER,
        'PWMDutyCycle-Mon': _c.V_DUTY_CYCLE,
        'CurrentInputIIBModIS-Mon': _c.V_I_INPUT_IS_IIB,
        'VoltageInputIIBModIS-Mon': _c.V_V_INPUT_IS_IIB,
        'IGBTTemperatureIIBModIS-Mon': _c.V_TEMP_IGBT_IS_IIB,
        'IGBTDriverVoltageIIBModIS-Mon': _c.V_V_DRIVER_IS_IIB,
        'IGBTDriverCurrentIIBModIS-Mon': _c.V_I_DRIVER_IS_IIB,
        'InductorTemperatureIIBModIS-Mon': _c.V_TEMP_INDUCTOR_IS_IIB,
        'HeatSinkTemperatureIIBModIS-Mon': _c.V_TEMP_HEATSINK_IS_IIB,
        'BoardTemperatureIIBModIS-Mon': _c.V_TEMP_BOARD_IS_IIB,
        'RelativeHumidityIIBModIS-Mon': _c.V_RH_IS_IIB,
        'IntlkIIBModIS-Mon': _c.V_IIB_INTERLOCKS_IS,
        'AlarmsIIBModIS-Mon': _c.V_IIB_ALARMS_IS,
        'VoltageOutputIIBModCmd-Mon': _c.V_V_OUTPUT_CMD_IIB,
        'CapacitorBankVoltageIIBModCmd-Mon': _c.V_V_CAPBANK_CMD_IIB,
        'RectInductorTempIIBModCmd-Mon': _c.V_TEMP_RECT_INDUCTOR_CMD_IIB,
        'RectHeatSinkTempIIBModCmd-Mon': _c.V_TEMP_RECT_HEATSINK_CMD_IIB,
        'ExtBoardsVoltageIIBModCmd-Mon': _c.V_V_EXT_BOARDS_CMD_IIB,
        'AuxBoardCurrentIIBModCmd-Mon': _c.V_I_AUX_BOARD_CMD_IIB,
        'IDBBoardCurrentIIBModCmd-Mon': _c.V_I_IDB_BOARD_CMD_IIB,
        'LeakCurrentIIBModCmd-Mon': _c.V_I_LEAKAGE_CMD_IIB,
        'BoardTemperatureIIBModCmd-Mon': _c.V_TEMP_BOARD_CMD_IIB,
        'RelativeHumidityIIBModCmd-Mon': _c.V_RH_CMD_IIB,
        'IntlkIIBModCmd-Mon': _c.V_IIB_INTERLOCKS_CMD,
        'AlarmsIIBModCmd-Mon': _c.V_IIB_ALARMS_CMD,
        }

    def writer(self, device_ids, epics_field, pru_controller, setpoints):
        """Return writer."""
        _c = PSModelFAC_2S_ACDC._c
        if epics_field == 'PwrState-Sel':
            return _writers.PSPwrState(device_ids, pru_controller, setpoints)
        elif epics_field == 'OpMode-Sel':
            return _writers.PSOpMode(
                device_ids,
                _writers.BSMPFunction(
                    device_ids, pru_controller, _c.F_SELECT_OP_MODE),
                setpoints)
        elif epics_field == 'CtrlLoop-Sel':
            return _writers.CtrlLoop(device_ids, pru_controller, setpoints)
        elif epics_field == 'CapacitorBankVoltage-SP':
            return _writers.BSMPFunction(
                device_ids, pru_controller, _c.F_SET_SLOWREF, setpoints)
        elif epics_field == 'Reset-Cmd':
            return _writers.Command(
                device_ids, pru_controller, _c.F_RESET_INTERLOCKS, setpoints)
        elif epics_field == 'Abort-Cmd':
            return _writers.BSMPFunctionNull()
        else:
            return _writers.BSMPFunctionNull()

    def controller(self, readers, writers, pru_controller, devices):
        """Return controller."""
        return _controller.PSController(
            readers, writers, pru_controller, devices)


class PSModelFAC_2P4S_ACDC(PSModelFAC_2S_ACDC):
    """FAC_2P4S_ACDC model."""

    _n = 'FAC_2P4S_ACDC'
    _c = _const_psbsmp.ConstFAC_2P4S_ACDC
    _e = _etity_psbsmp.EntitiesFAC_2P4S_ACDC

    _bsmp_variables = {
        'CapacitorBankVoltage-RB': _c.V_PS_SETPOINT,
        'CapacitorBankVoltageRef-Mon': _c.V_PS_REFERENCE,
        'IntlkSoft-Mon': _c.V_PS_SOFT_INTERLOCKS,
        'IntlkHard-Mon': _c.V_PS_HARD_INTERLOCKS,
        'CapacitorBankVoltage-Mon': _c.V_V_CAPACITOR_BANK,
        'RectifierCurrent-Mon': _c.V_I_OUT_RECTIFIER,
        'PWMDutyCycle-Mon': _c.V_DUTY_CYCLE,
        'CurrentInputIIBModIS-Mon': _c.V_I_INPUT_IS_IIB,
        'VoltageInputIIBModIS-Mon': _c.V_V_INPUT_IS_IIB,
        'IGBTTemperatureIIBModIS-Mon': _c.V_TEMP_IGBT_IS_IIB,
        'IGBTDriverVoltageIIBModIS-Mon': _c.V_V_DRIVER_IS_IIB,
        'IGBTDriverCurrentIIBModIS-Mon': _c.V_I_DRIVER_IS_IIB,
        'InductorTemperatureIIBModIS-Mon': _c.V_TEMP_INDUCTOR_IS_IIB,
        'HeatSinkTemperatureIIBModIS-Mon': _c.V_TEMP_HEATSINK_IS_IIB,
        'BoardTemperatureIIBModIS-Mon': _c.V_TEMP_BOARD_IS_IIB,
        'RelativeHumidityIIBModIS-Mon': _c.V_RH_IS_IIB,
        'IntlkIIBModIS-Mon': _c.V_IIB_INTERLOCKS_IS,
        'AlarmsIIBModIS-Mon': _c.V_IIB_ALARMS_IS,
        'VoltageOutputIIBModCmd-Mon': _c.V_V_OUTPUT_CMD_IIB,
        'CapacitorBankVoltageIIBModCmd-Mon': _c.V_V_CAPBANK_CMD_IIB,
        'InductorTemperatureIIBModCmd-Mon': _c.V_TEMP_RECT_INDUCTOR_CMD_IIB,
        'HeatSinkTemperatureIIBModCmd-Mon': _c.V_TEMP_RECT_HEATSINK_CMD_IIB,
        'ExtBoardsVoltageIIBModCmd-Mon': _c.V_V_EXT_BOARDS_CMD_IIB,
        'AuxBoardCurrentIIBModCmd-Mon': _c.V_I_AUX_BOARD_CMD_IIB,
        'IDBBoardCurrentIIBModCmd-Mon': _c.V_I_IDB_BOARD_CMD_IIB,
        'LeakCurrentIIBModCmd-Mon': _c.V_I_LEAKAGE_CMD_IIB,
        'BoardTemperatureIIBModCmd-Mon': _c.V_TEMP_BOARD_CMD_IIB,
        'RelativeHumidityIIBModCmd-Mon': _c.V_RH_CMD_IIB,
        'IntlkIIBModCmd-Mon': _c.V_IIB_INTERLOCKS_CMD,
        'AlarmsIIBModCmd-Mon': _c.V_IIB_ALARMS_CMD,
        }


# --- Factory ---


class PSModelFactory:
    """PSModel Factory."""

    _psname_2_factory = {
        'FBP': PSModelFBP,
        'FBP_DCLink': PSModelFBP_DCLink,
        'FAC_DCDC': PSModelFAC_DCDC,
        'FAC_2S_DCDC': PSModelFAC_2S_DCDC,
        'FAC_2S_ACDC': PSModelFAC_2S_ACDC,
        'FAC_2P4S_DCDC': PSModelFAC_2P4S_DCDC,
        'FAC_2P4S_ACDC': PSModelFAC_2P4S_ACDC,
        'FAP': PSModelFAP,
        'FAP_2P2S': PSModelFAP_2P2S,
        'FAP_4P': PSModelFAP_4P,
        }

    @staticmethod
    def create(psmodel):
        """Return PSModel object."""
        if psmodel in PSModelFactory._psname_2_factory:
            factory = PSModelFactory._psname_2_factory[psmodel]
            return factory()
        raise ValueError('PS Model "{}" not defined'.format(psmodel))



"""PVs Configs Handler."""

import time as _time
import numpy as _np
from epics import get_pv as _get_pv

from . import ConfigDBDocument as _ConfigDBDocument


_TIMEOUT = 0.5


class PVsConfig(_ConfigDBDocument):
    """Class to handle PVs configurations."""

    PVs = dict()

    def __init__(self, config_type='global_config', name=None, url=None):
        """Init."""
        super().__init__(config_type, name=name, url=url)

    def connect(self):
        """Create PVs."""
        template = self.get_value_from_template()
        for pvn, _, _ in template['pvs']:
            self._get_pv(pvn)

    @property
    def connected(self):
        """."""
        template = self.get_value_from_template()
        for pvn, _, _ in template['pvs']:
            pvobj = self.PVs.get(pvn)
            if pvobj is None:
                return False
            elif not pvobj.connected:
                return False
        return True

    @property
    def pvs(self):
        """Return dict with PVs as keys and (values, delay) as value."""
        if self._value is None:
            pvslist = self.get_value_from_template()
        else:
            pvslist = self._value['pvs']
        return {item[0]: item[1:] for item in pvslist}

    def read(self, timeout=_TIMEOUT):
        """Read machine state."""
        new_config_value = dict()
        new_config_value['pvs'] = list()

        template = self.get_value_from_template()

        # connect
        self.connect()

        # read
        pvs_not_read = set()
        for pvn, defval, delay in template['pvs']:
            pvobj = self._get_pv(pvn)
            if pvobj.wait_for_connection(timeout):
                value = defval if pvn.endswith('-Cmd')\
                    else pvobj.get(timeout=timeout)
            else:
                pvs_not_read.add(pvn)
                value = 0
            new_config_value['pvs'].append([pvn, value, delay])

        if pvs_not_read:
            return False, pvs_not_read

        self.value = new_config_value
        return True, []

    def read_and_save(self, new_name=None):
        """Read machine state and save new configuration in config server."""
        status_ok, failed_list = self.read()
        if not status_ok:
            return False, failed_list
        self.save(new_name)
        return True, []

    def apply(self, pvsdict=None, timeout=_TIMEOUT):
        """Apply current config value to machine and check if implemented."""
        if not pvsdict:
            pvsdict = self.pvs

        # connect
        self.connect()

        pvs_not_set = set()

        # set
        for pvn, (value, delay) in pvsdict.items():
            pvobj = self._get_pv(pvn)
            try:
                pvobj.put(value)
                _time.sleep(delay)
            except TypeError:
                pvs_not_set.add(pvn)

        # wait
        _time.sleep(2.0)

        # check
        for pvn, (value, delay) in pvsdict.items():
            if pvn in pvs_not_set:
                continue
            equal = self._check_pv(pvn, value, timeout=timeout)
            if not equal:
                pvs_not_set.add(pvn)

        if pvs_not_set:
            return False, pvs_not_set
        return True, []

    def load_and_apply(self, timeout=_TIMEOUT, discarded=False):
        """Load from server, apply to machine and check if implemented."""
        self.load(discarded=discarded)
        status_ok, failed_list = self.apply(timeout=timeout)
        if not status_ok:
            return False, failed_list
        return True, []

    # ---------- private methods ----------

    def _get_pv(self, pvname, timeout=_TIMEOUT):
        """Return PV object."""
        pvobj = self.PVs.get(pvname)
        if pvobj is None:
            pvobj = _get_pv(pvname, timeout=timeout)
            self.PVs[pvname] = pvobj
        return pvobj

    def _check_pv(
            self, pvname, value, timeout=_TIMEOUT, rel_tol=1e-06, abs_tol=0.0):
        """Check PV value."""
        pvobj = self._get_pv(pvname)
        pvobj.wait_for_connection(timeout)
        curr_val = pvobj.get(timeout=timeout)
        if curr_val is None:
            return False
        elif isinstance(curr_val, (_np.ndarray, list, tuple)) or \
                isinstance(value, (_np.ndarray, list, tuple)):
            try:
                if len(curr_val) != len(value):
                    return False
            except TypeError:
                return False  # one of them is not an array
            return _np.allclose(curr_val, value, rtol=rel_tol, atol=abs_tol)
        elif isinstance(curr_val, float) or isinstance(value, float):
            return _np.isclose(curr_val, value, rtol=rel_tol, atol=abs_tol)
        elif curr_val == value:
            return True
        return False



"""FOFB acquisition devices."""

import time as _time
import numpy as _np

from ..search import PSSearch as _PSSearch, BPMSearch as _BPMSearch
from ..csdev import Const as _Const
from ..fofb.csdev import NR_BPM

from .device import ProptyDevice as _ProptyDevice, Devices as _Devices, \
    Device as _Device
from .fofb import FOFBCtrlBase as _FOFBCtrlBase
from .timing import Event
from .psconv import StrengthConv


class _FOFBCtrlAcqConst(_Const):
    """FOFB acquisition constants."""

    # TODO: when BPM IOC is updated, these acquisition core base classes
    # should be migrated to a common module to be shared.

    CHANNEL = ('lamp', 'invalid', 'sysid', 'sysid_applied')
    TRIGTYP = ('now', 'external', 'data', 'software')
    TRIGEVT = ('start', 'stop')
    REPEAT = ('normal', 'repetitive')
    STATES = (
        'Idle', 'Waiting', 'External Trig', 'Data Trig', 'Software Trig',
        'Acquiring', 'Error', 'Bad Post Samples', 'Too Many Samples',
        'No Samples', 'No Memory', 'Overflow')

    Channel = _Const.register('Channel', CHANNEL)
    TrigTyp = _Const.register('TrigTyp', TRIGTYP)
    TrigEvt = _Const.register('Events', TRIGEVT)
    Repeat = _Const.register('Repeat', REPEAT)
    States = _Const.register('States', STATES)

    STATES_NOTOK = {
        States.Error, States.No_Memory, States.Bad_Post_Samples,
        States.No_Samples, States.Too_Many_Samples, States.Overflow}
    STATES_STARTED = {
        States.Waiting, States.External_Trig, States.Data_Trig,
        States.Software_Trig, States.Acquiring}
    STATES_FINISHED = {States.Idle}
    STATES_FINISHED |= STATES_NOTOK


class _FOFBCtrlAcqBase(_ProptyDevice, _FOFBCtrlBase, _FOFBCtrlAcqConst):
    """FOFB acquisition base device."""

    _properties = (
        'Channel-Sel', 'Channel-Sts',
        'Shots-SP', 'Shots-RB',
        'UpdateTime-SP', 'UpdateTime-RB',
        'TriggerRep-Sel', 'TriggerRep-Sts',
        'SamplesPre-SP', 'SamplesPre-RB',
        'SamplesPost-SP', 'SamplesPost-RB',
        'Trigger-Sel', 'Trigger-Sts',
        'TriggerEvent-Cmd', 'Status-Mon', 'Count-Mon',
    )

    def __init__(self, devname, **kws):
        """Init."""
        # call base class constructor
        super().__init__(devname, **kws)

    @property
    def channel(self):
        """Channel enum index."""
        return self['Channel-Sts']

    @channel.setter
    def channel(self, val):
        self['Channel-Sel'] = val

    @property
    def channel_str(self):
        """Channel enum string."""
        return _FOFBCtrlAcqConst.CHANNEL[self.channel]

    @property
    def nrshots(self):
        """Number of shots in the acquisition."""
        return self['Shots-RB']

    @nrshots.setter
    def nrshots(self, val):
        self['Shots-SP'] = val

    @property
    def update_time(self):
        """Acquisition update time."""
        return self['UpdateTime-RB']  # / 1e3

    @update_time.setter
    def update_time(self, val):
        self['UpdateTime-SP'] = val  # * 1e3

    @property
    def repeat(self):
        """Whether the acquisition is repeat or normal."""
        return self['TriggerRep-Sts']

    @repeat.setter
    def repeat(self, val):
        self['TriggerRep-Sel'] = val

    @property
    def nrsamples_pre(self):
        """Number of samples previous to trigger acquisition."""
        return self['SamplesPre-RB']

    @nrsamples_pre.setter
    def nrsamples_pre(self, val):
        self['SamplesPre-SP'] = val

    @property
    def nrsamples_post(self):
        """Number of samples posterior to trigger acquisition."""
        return self['SamplesPost-RB']

    @nrsamples_post.setter
    def nrsamples_post(self, val):
        self['SamplesPost-SP'] = val

    @property
    def trigger(self):
        """Trigger."""
        return self['Trigger-Sts']

    @trigger.setter
    def trigger(self, val):
        self['Trigger-Sel'] = val

    @property
    def status(self):
        """Acquisition status."""
        return self['Status-Sts']

    @property
    def count(self):
        """Counter of number of acquisitions so far."""
        return self['Count-Mon']

    def cmd_ctrl(self, val):
        """Command to start or stop acquisition."""
        self['TriggerEvent-Cmd'] = val

    def wait_acq_finish(self, timeout=10):
        """Wait Acquisition to finish."""
        return self._wait(
            'Status-Sts', _FOFBCtrlAcqConst.STATES_FINISHED,
            timeout=timeout, comp=lambda x, y: x in y)

    def wait_acq_start(self, timeout=10):
        """Wait Acquisition to start."""
        return self._wait(
            'Status-Sts', _FOFBCtrlAcqConst.STATES_STARTED,
            timeout=timeout, comp=lambda x, y: x in y)


class _FamFOFBAcqBase(_Devices, _FOFBCtrlAcqConst):

    DEF_TIMEOUT = 10  # [s]
    FOFBCTRL_CLASS = _FOFBCtrlAcqBase
    FOFBPS_CLASS = _FOFBCtrlAcqBase

    def __init__(self, ctrlname=None, psnames=None, bpmnames=None, event=None):
        # FOFBCtrl devices
        ctrlnames = _FOFBCtrlBase.DEVICES if not ctrlname else [ctrlname, ]
        self._ctlrs, subs = dict(), list()
        for idx, ctl in enumerate(ctrlnames):
            self._ctlrs[ctl] = self.FOFBCTRL_CLASS(ctl, auto_monitor_mon=False)
            subs.append(f'{idx+1:02}')
        sub = '(' + '|'.join(subs) + ').*'

        # FOFBPS devices
        if psnames is None:
            chn = _PSSearch.get_psnames(dict(sec='SI', sub=sub, dev='FCH'))
            cvn = _PSSearch.get_psnames(dict(sec='SI', sub=sub, dev='FCV'))
            psnames = chn + cvn
        self._psnames = psnames
        self._psdevs = {psn: self.FOFBPS_CLASS(psn, auto_monitor_mon=False)
                        for psn in self._psnames}
        self._psconv = {psn: StrengthConv(psn, 'Ref-Mon')
                        for psn in self._psnames}

        # BPM names
        self._bpm_names = bpmnames or _BPMSearch.get_names(
            dict(sec='SI', sub=sub, dev='BPM'))

        # FOFBS event
        event = event or 'FOFBS'
        self._evt_fofb = Event(event)

        # call base class constructor
        devices = list()
        devices.extend(self._ctlrs.values())
        devices.extend(self._psdevs.values())
        devices.extend(self._psconv.values())
        super().__init__('SI-Glob:BS-FOFBAcq', devices)

    @property
    def ctrldevs(self):
        """FOFBCtrl device list."""
        return self._ctlrs

    @property
    def psnames(self):
        """PS name list."""
        return list(self._psnames)

    @property
    def psdevs(self):
        """PS device list."""
        return self._psdevs

    @property
    def psconvs(self):
        """PS conversion device list."""
        return self._psconv

    @property
    def bpmnames(self):
        """BPM names."""
        return self._bpm_names

    @property
    def evtdev(self):
        """FOFBS event device."""
        return self._evt_fofb

    @property
    def channel(self):
        """Channel enum index for all controllers in ctrldevs."""
        return _np.array([dev.channel for dev in self._ctlrs.values()])

    @property
    def nrshots(self):
        """
        Number of shots in the acquisition for all
        controllers in ctrldevs.
        """
        return _np.array([dev.nrshots for dev in self._ctlrs.values()])

    @property
    def update_time(self):
        """Acquisition update time for all controllers in ctrldevs."""
        return _np.array([dev.update_time for dev in self._ctlrs.values()])

    @property
    def repeat(self):
        """
        Whether the acquisition is repeat or normal for all
        controllers in ctrldevs.
        """
        return _np.array([dev.repeat for dev in self._ctlrs.values()])

    @property
    def nrsamples_pre(self):
        """
        Number of samples previous to trigger acquisition for all
        controllers in ctrldevs.
        """
        return _np.array([dev.nrsamples_pre for dev in self._ctlrs.values()])

    @property
    def nrsamples_post(self):
        """
        Number of samples posterior to trigger acquisition for all
        controllers in ctrldevs.
        """
        return _np.array([dev.nrsamples_post for dev in self._ctlrs.values()])

    @property
    def trigger(self):
        """Trigger source for all controllers in ctrldevs."""
        return _np.array([dev.trigger for dev in self._ctlrs.values()])

    @property
    def status(self):
        """Acquisition status."""
        data = list()
        for dev in self._ctlrs.values():
            data.append(dev.status)

        if len(data) == 1:
            return data[0]
        return data

    def set_channel(self, val):
        """Set channel for all FOFBCtrl."""
        self._set_devices_propty(self._ctlrs, 'Channel-Sel', val)
        return self._wait_devices_propty(self._ctlrs, 'Channel-Sts', val)

    def set_nrshots(self, val):
        """Set nrshots for all FOFBCtrl."""
        self._set_devices_propty(self._ctlrs, 'Shots-SP', val)
        return self._wait_devices_propty(self._ctlrs, 'Shots-RB', val)

    def set_update_time(self, val):
        """Set update_time for all FOFBCtrl."""
        self._set_devices_propty(self._ctlrs, 'UpdateTime-SP', val)
        return self._wait_devices_propty(self._ctlrs, 'UpdateTime-RB', val)

    def set_repeat(self, val):
        """Set repeat for all FOFBCtrl."""
        self._set_devices_propty(self._ctlrs, 'TriggerRep-Sel', val)
        return self._wait_devices_propty(self._ctlrs, 'TriggerRep-Sts', val)

    def set_nrsamples_pre(self, val):
        """Set nrsamples_pre for all FOFBCtrl."""
        self._set_devices_propty(self._ctlrs, 'SamplesPre-SP', val)
        return self._wait_devices_propty(self._ctlrs, 'SamplesPre-RB', val)

    def set_nrsamples_post(self, val):
        """Set nrsamples_post for all FOFBCtrl."""
        self._set_devices_propty(self._ctlrs, 'SamplesPost-SP', val)
        return self._wait_devices_propty(self._ctlrs, 'SamplesPost-RB', val)

    def set_trigger(self, val):
        """Set trigger for all FOFBCtrl."""
        self._set_devices_propty(self._ctlrs, 'Trigger-Sel', val)
        return self._wait_devices_propty(self._ctlrs, 'Trigger-Sts', val)


# ------- system identification devices -------

class FOFBCtrlSysId(_FOFBCtrlAcqBase):
    """FOFB controller system identification device."""

    _properties = _FOFBCtrlAcqBase._properties
    _properties += ('TimeFrameData', 'DataType')
    _properties += tuple([f'BPM{i}PosXData' for i in range(8)])
    _properties += tuple([f'BPM{i}PosYData' for i in range(8)])
    _properties += (
        'PRBSSyncEn-Sel', 'PRBSSyncEn-Sts',
        'PRBSStepDuration-SP', 'PRBSStepDuration-RB',
        'PRBSLFSRLength-SP', 'PRBSLFSRLength-RB',
        'PRBSFOFBAccMovAvgTaps-SP', 'PRBSFOFBAccMovAvgTaps-RB',
        'PRBSFOFBAccEn-Sel', 'PRBSFOFBAccEn-Sts',
        'PRBSBPMPosEn-Sel', 'PRBSBPMPosEn-Sts',
        'PRBSBPMPosXLvl0-SP', 'PRBSBPMPosXLvl0-RB',
        'PRBSBPMPosXLvl1-SP', 'PRBSBPMPosXLvl1-RB',
        'PRBSBPMPosYLvl0-SP', 'PRBSBPMPosYLvl0-RB',
        'PRBSBPMPosYLvl1-SP', 'PRBSBPMPosYLvl1-RB',
        'PRBSData',
    )

    def __init__(self, devname, auto_monitor_mon=True):
        """Init."""
        # call base class constructor
        super().__init__(
            devname, propty_prefix='SYSID',
            properties=FOFBCtrlSysId._properties,
            auto_monitor_mon=auto_monitor_mon)

    @property
    def data_type(self):
        """Data type."""
        return self['DataType']

    @data_type.setter
    def data_type(self, val):
        self['DataType'] = val

    @property
    def timeframe_data(self):
        """Time frame data."""
        return self['TimeFrameData']

    def get_bpm_posx(self, bpm_index):
        """Return BPM PosXData for BPM of index bpm_index."""
        return self[f'BPM{bpm_index}PosXData']

    def get_bpm_posy(self, bpm_index):
        """Return BPM PosYData for BPM of index bpm_index."""
        return self[f'BPM{bpm_index}PosYData']

    @property
    def prbs_data(self):
        """PRBS Data."""
        return self['PRBSData']

    @property
    def prbs_sync_enbl(self):
        """PRBS sync enable."""
        return self['PRBSSyncEn-Sts']

    @prbs_sync_enbl.setter
    def prbs_sync_enbl(self, val):
        self['PRBSSyncEn-Sel'] = val

    @property
    def prbs_step_duration(self):
        """PRBS step duration."""
        return self['PRBSStepDuration-RB']

    @prbs_step_duration.setter
    def prbs_step_duration(self, val):
        self['PRBSStepDuration-SP'] = val

    @property
    def prbs_lfsr_len(self):
        """PRBS LFSR length."""
        return self['PRBSLFSRLength-RB']

    @prbs_lfsr_len.setter
    def prbs_lfsr_len(self, val):
        self['PRBSLFSRLength-SP'] = val

    @property
    def prbs_mov_avg_taps(self):
        """Number of taps of the PRBS moving average filter."""
        return self['PRBSFOFBAccMovAvgTaps-RB']

    @prbs_mov_avg_taps.setter
    def prbs_mov_avg_taps(self, val):
        self['PRBSFOFBAccMovAvgTaps-SP'] = val

    @property
    def prbs_fofbacc_enbl(self):
        """Whether to enable or not PRBS actuation in correctors."""
        return self['PRBSFOFBAccEn-Sts']

    @prbs_fofbacc_enbl.setter
    def prbs_fofbacc_enbl(self, val):
        self['PRBSFOFBAccEn-Sel'] = val

    @property
    def prbs_bpmpos_enbl(self):
        """Whether to enable or not PRBS excitation in BPMs PosX."""
        return self['PRBSBPMPosEn-Sts']

    @prbs_bpmpos_enbl.setter
    def prbs_bpmpos_enbl(self, val):
        self['PRBSBPMPosEn-Sel'] = val

    @property
    def prbs_bpmposx_lvl0(self):
        """Array with all BPMs PosX value for PRBS level 0  [nm]."""
        return self['PRBSBPMPosXLvl0-RB']

    @prbs_bpmposx_lvl0.setter
    def prbs_bpmposx_lvl0(self, val):
        self['PRBSBPMPosXLvl0-SP'] = _np.array(val, dtype=int)

    @property
    def prbs_bpmposx_lvl1(self):
        """Array with all BPMs PosX value for PRBS level 1  [nm]."""
        return self['PRBSBPMPosXLvl1-RB']

    @prbs_bpmposx_lvl1.setter
    def prbs_bpmposx_lvl1(self, val):
        self['PRBSBPMPosXLvl1-SP'] = _np.array(val, dtype=int)

    @property
    def prbs_bpmposy_lvl0(self):
        """Array with all BPMs PosY value for PRBS level 0  [nm]."""
        return self['PRBSBPMPosYLvl0-RB']

    @prbs_bpmposy_lvl0.setter
    def prbs_bpmposy_lvl0(self, val):
        self['PRBSBPMPosYLvl0-SP'] = _np.array(val, dtype=int)

    @property
    def prbs_bpmposy_lvl1(self):
        """Array with all BPMs PosY value for PRBS level 1  [nm]."""
        return self['PRBSBPMPosYLvl1-RB']

    @prbs_bpmposy_lvl1.setter
    def prbs_bpmposy_lvl1(self, val):
        self['PRBSBPMPosYLvl1-SP'] = _np.array(val, dtype=int)


class FOFBPSSysId(_Device):
    """FOFB power supply system identification device."""

    _properties = (
        'SYSIDFOFBAccRawData', 'SYSIDFOFBAccData',
        'SYSIDPRBSFOFBAccLvl0-SP', 'SYSIDPRBSFOFBAccLvl0-RB',
        'SYSIDPRBSFOFBAccLvl1-SP', 'SYSIDPRBSFOFBAccLvl1-RB',
        'CurrLoopKp-SP', 'CurrLoopKp-RB',
        'CurrLoopTi-SP', 'CurrLoopTi-RB',
    )

    def __init__(self, devname, auto_monitor_mon=True):
        """Init."""
        # call base class constructor
        super().__init__(
            devname, properties=FOFBPSSysId._properties,
            auto_monitor_mon=auto_monitor_mon)

    @property
    def fofbacc_rawdata(self):
        """FOFB accumulator raw data for SYSID core."""
        return self['SYSIDFOFBAccRawData']

    @property
    def fofbacc_data(self):
        """FOFB accumulator data for SYSID core."""
        return self['SYSIDFOFBAccData']

    @property
    def prbs_fofbacc_lvl0(self):
        """Actuation value for level 0 of SYSID PRBS signal."""
        return self['SYSIDPRBSFOFBAccLvl0-RB']

    @prbs_fofbacc_lvl0.setter
    def prbs_fofbacc_lvl0(self, val):
        self['SYSIDPRBSFOFBAccLvl0-SP'] = val

    @property
    def prbs_fofbacc_lvl1(self):
        """Actuation value for level 1 of SYSID PRBS signal."""
        return self['SYSIDPRBSFOFBAccLvl1-RB']

    @prbs_fofbacc_lvl1.setter
    def prbs_fofbacc_lvl1(self, val):
        self['SYSIDPRBSFOFBAccLvl1-SP'] = val

    @property
    def currloop_kp(self):
        """Current control loop Kp parameter."""
        return self['CurrLoopKp-RB']

    @currloop_kp.setter
    def currloop_kp(self, value):
        self['CurrLoopKp-SP'] = value

    @property
    def currloop_ti(self):
        """Current control loop Ti parameter."""
        return self['CurrLoopTi-RB']

    @currloop_ti.setter
    def currloop_ti(self, value):
        self['CurrLoopTi-SP'] = value


class FamFOFBSysId(_FamFOFBAcqBase):
    """Family of FOFB SYSID acquisition device."""

    FOFBCTRL_CLASS = FOFBCtrlSysId
    FOFBPS_CLASS = FOFBPSSysId
    DEF_ATOL_FOFBACC = 1e-6

    def __init__(self, **kws):
        super().__init__(**kws)
        self._initial_timestamps = None

    @property
    def data_type(self):
        """Data type of all controllers in ctrldevs."""
        return _np.array(
            [dev.data_type for dev in self._ctlrs.values()])

    @property
    def timeframe_data(self):
        """Time frame data of all controllers in ctrldevs."""
        return [dev.timeframe_data for dev in self._ctlrs.values()]

    @property
    def prbs_data(self):
        """PRBS Data of all controllers in ctrldevs."""
        return [dev.prbs_data for dev in self._ctlrs.values()]

    @property
    def prbs_step_duration(self):
        """PRBS step duration of all controllers in ctrldevs."""
        return _np.array(
            [dev.prbs_step_duration for dev in self._ctlrs.values()])

    @property
    def prbs_lfsr_len(self):
        """PRBS LFSR length of all controllers in ctrldevs."""
        return _np.array(
            [dev.prbs_lfsr_len for dev in self._ctlrs.values()])

    @property
    def prbs_mov_avg_taps(self):
        """
        Number of taps of the PRBS moving average filter of all controllers.
        """
        return _np.array(
            [dev.prbs_mov_avg_taps for dev in self._ctlrs.values()])

    @property
    def prbs_sync_enbl(self):
        """PRBS sync enable of all controllers in ctrldevs."""
        return _np.array(
            [dev.prbs_sync_enbl for dev in self._ctlrs.values()])

    @property
    def prbs_fofbacc_enbl(self):
        """
        Whether to enable or not PRBS actuation in correctors
        of all controllers in ctrldevs.
        """
        return _np.array(
            [dev.prbs_fofbacc_enbl for dev in self._ctlrs.values()])

    @property
    def prbs_fofbacc_lvl0(self):
        """Array with correctors actuation value for PRBS level 0 [A]."""
        return _np.array(
            [self._psdevs[psn].prbs_fofbacc_lvl0 for psn in self._psnames])

    @property
    def prbs_fofbacc_lvl1(self):
        """Array with correctors actuation value for PRBS level 1 [A]."""
        return _np.array(
            [self._psdevs[psn].prbs_fofbacc_lvl1 for psn in self._psnames])

    @property
    def prbs_bpmpos_enbl(self):
        """
        Whether to enable or not PRBS excitation in BPMs PosX
        of all controllers in ctrldevs.
        """
        return _np.array(
            [dev.prbs_bpmpos_enbl for dev in self._ctlrs.values()])

    @property
    def prbs_bpmposx_lvl0(self):
        """
        Array with all BPMs PosX value for PRBS level 0 of all
        controllers in ctrldevs [nm].
        """
        return _np.array(
            [dev.prbs_bpmposx_lvl0 for dev in self._ctlrs.values()])

    @property
    def prbs_bpmposx_lvl1(self):
        """
        Array with all BPMs PosX value for PRBS level 1 of all
        controllers in ctrldevs [nm].
        """
        return _np.array(
            [dev.prbs_bpmposx_lvl1 for dev in self._ctlrs.values()])

    @property
    def prbs_bpmposy_lvl0(self):
        """
        Array with all BPMs PosY value for PRBS level 0 of all
        controllers in ctrldevs [nm].
        """
        return _np.array(
            [dev.prbs_bpmposy_lvl0 for dev in self._ctlrs.values()])

    @property
    def prbs_bpmposy_lvl1(self):
        """
        Array with all BPMs PosY value for PRBS level 1 of all
        controllers in ctrldevs [nm].
        """
        return _np.array(
            [dev.prbs_bpmposy_lvl1 for dev in self._ctlrs.values()])

    @property
    def currloop_kp(self):
        """Current Loop Kp.

        Returns:
            kp (numpy.ndarray, 160):
                current loop Kp for each power supply.
        """
        return _np.array(
            [self._psdevs[psn].currloop_kp for psn in self._psnames])

    @property
    def currloop_ti(self):
        """Current Loop Ti.

        Returns:
            ti (numpy.ndarray, 160):
                current loop Ti for each power supply.
        """
        return _np.array(
            [self._psdevs[psn].currloop_ti for psn in self._psnames])

    @property
    def strength_2_current_factor(self):
        """Strength to current convertion factor.

        Returns:
            factor (numpy.ndarray, NR_BPM):
                convertion factor for each power supply.
        """
        return _np.array(
            [self._psconv[psn].conv_strength_2_current(1, strengths_dipole=3.0)
             for psn in self._psnames])

    def set_data_type(self, value):
        """Configure acquisition data type for FOFB controllers.

        Args:
            data_type (int): acquisition data type
                (0:raw, 1:lamp, 2:sysid).
        """
        for ctl in self._ctlrs.values():
            ctl.data_type = value
        return True

    def check_data_valid(self):
        """Check whether acquisition TimeFrameData are valid.

        Returns:
            int: code describing what happened:
                -1: TimeFrameData is not monotonic;
                =0: data is valid.
                >0: index of the first FOFB controller which have time frame
                    or PRBS data different from controller 01.

        """
        # check whether time frame data are synced between controllers
        ctlr01_tfdata = self.timeframe_data[0]
        for idx, data in enumerate(self.timeframe_data):
            if not _np.all(data == ctlr01_tfdata):
                return idx+1
        # check if timeframe data is monotonic
        if not all(v in (1, -2**16+1) for v in _np.diff(ctlr01_tfdata)):
            return -1
        # check whether prbs data are synced between controllers
        ctlr01_prbsdata = self.prbs_data[0]
        for idx, data in enumerate(self.prbs_data):
            if not _np.all(data == ctlr01_prbsdata):
                return idx+1
        return 0

    def set_prbs_mov_avg_taps(self, value):
        """Configure number of taps of the PRBS moving average filter
        for all FOFB controllers.

        Args:
            value (int): number of taps.
        """
        for ctl in self._ctlrs.values():
            ctl.prbs_mov_avg_taps = value
        return True

    def config_prbs(self, step_duration, lfsr_len):
        """Configure PRBS for FOFB controllers.

        Args:
            step_duration (int): step duration.
            lfsr_len (int): LFSR length.
        """
        for ctl in self._ctlrs.values():
            ctl.prbs_step_duration = step_duration
            ctl.prbs_lfsr_len = lfsr_len
        return True

    def cmd_sync_prbs(self, timeout=_FamFOFBAcqBase.DEF_TIMEOUT):
        """Command to sync PRBS signal for all controllers."""
        ctrls = list(self._ctlrs.values())

        self._set_devices_propty(ctrls, 'PRBSSyncEn-Sel', 1)
        if not self._wait_devices_propty(
                ctrls, 'PRBSSyncEn-Sts', 1, timeout=timeout/2):
            return False

        self._evt_fofb.cmd_external_trigger()
        _time.sleep(2)

        self._set_devices_propty(ctrls, 'PRBSSyncEn-Sel', 0)
        if not self._wait_devices_propty(
                ctrls, 'PRBSSyncEn-Sts', 0, timeout=timeout/2):
            return False
        return True

    def set_prbs_fofbacc_levels(self, level0, level1=None):
        """Set power supply FOFBAcc PRBS excitation levels.

        Args:
            level0 (numpy.ndarray, (NR_CORRS,)):
                desired values for power supplies PRBS excitation level 0
                in psnames order.
            level1 (optional, numpy.ndarray, (NR_CORRS,)):
                If None, we consider level1 = -level0. Defaults to None.

        """
        level0, level1 = self._handle_corr_levels(level0, level1)
        for i, psn in enumerate(self._psnames):
            dev = self._psdevs[psn]
            dev.prbs_fofbacc_lvl0 = level0[i]
            dev.prbs_fofbacc_lvl1 = level1[i]
        return True

    def check_prbs_fofbacc_levels(
            self, level0, level1=None, atol=DEF_ATOL_FOFBACC):
        """
        Check whether power supply FOFBAcc PRBS excitation levels are equal to
        desired values.

        Args:
            level0 (numpy.ndarray, (NR_CORRS,)):
                desired values for power supplies PRBS excitation level 0
                in psnames order.
            level1 (optional, numpy.ndarray, (NR_CORRS,)):
                If None, we consider level1 = -level0. Defaults to None.
            atol (optional, float):
                Absolute tolerance for values comparison.

        Returns:
            bool: indicate whether levels 0 and 1 of power supply FOFBAcc PRBS
                excitation are in desired values.

        """
        level0, level1 = self._handle_corr_levels(level0, level1)
        impltd0 = _np.array(
            [self._psdevs[psn].prbs_fofbacc_lvl0 for psn in self._psnames])
        if not _np.allclose(level0, impltd0, atol=atol):
            return False
        impltd1 = _np.array(
            [self._psdevs[psn].prbs_fofbacc_lvl1 for psn in self._psnames])
        if not _np.allclose(level1, impltd1, atol=atol):
            return False
        return True

    def wait_prbs_fofbacc_levels(
            self, level0, level1=None, atol=DEF_ATOL_FOFBACC,
            timeout=_FamFOFBAcqBase.DEF_TIMEOUT):
        """
        Wait for power supply FOFBAcc PRBS excitation levels to be equal to
        desired values. Refer to check_prbs_fofbacc_levels for more details
        on arguments and returns.
        """
        _t0 = _time.time()
        while _time.time() - _t0 < timeout:
            if self.check_prbs_fofbacc_levels(level0, level1, atol):
                return True
            _time.sleep(0.1)
        return False

    def _handle_corr_levels(self, level0, level1):
        nr_ps = len(self._psnames)
        if isinstance(level0, (int, float, bool)):
            level0 = nr_ps * [level0]
        level0 = _np.asarray(level0)
        if level1 is None:
            level1 = -1 * level0
        if len(level0) != nr_ps or len(level1) != nr_ps:
            raise ValueError('Levels and psnames must have the same length.')
        return level0, level1

    def cmd_prbs_fofbacc_enable(self, timeout=_FamFOFBAcqBase.DEF_TIMEOUT):
        """Command to enable PRBS for correctors."""
        ctrls = list(self._ctlrs.values())
        self._set_devices_propty(ctrls, 'PRBSFOFBAccEn-Sel', 1)
        if not self._wait_devices_propty(
                ctrls, 'PRBSFOFBAccEn-Sts', 1, timeout=timeout):
            return False
        return True

    def cmd_prbs_fofbacc_disable(self, timeout=_FamFOFBAcqBase.DEF_TIMEOUT):
        """Command to disable PRBS for correctors."""
        ctrls = list(self._ctlrs.values())
        self._set_devices_propty(ctrls, 'PRBSFOFBAccEn-Sel', 0)
        if not self._wait_devices_propty(
                ctrls, 'PRBSFOFBAccEn-Sts', 0, timeout=timeout/2):
            return False
        return True

    def set_prbs_bpmposx_levels(self, level0, level1=None):
        """Set BPM PosX PRBS excitation levels.

        Args:
            level0 (numpy.ndarray, (NR_BPM,)):
                desired values for BPM PosX PRBS excitation level 0.
            level1 (optional, numpy.ndarray, (NR_BPM,)):
                If None, we consider level1 = -level0. Defaults to None.

        """
        level0, level1 = self._handle_bpm_levels(level0, level1)
        for ctl in self._ctlrs.values():
            ctl.prbs_bpmposx_lvl0 = level0
            ctl.prbs_bpmposx_lvl1 = level1
        return True

    def check_prbs_bpmposx_levels(self, level0, level1=None):
        """
        Check whether BPM PosX PRBS excitation levels are equal to
        desired values.

        Args:
            level0 (numpy.ndarray, (NR_BPM,)):
                desired values for BPM PosX PRBS excitation level 0.
            level1 (optional, numpy.ndarray, (NR_BPM,)):
                If None, we consider level1 = -level0. Defaults to None.

        Returns:
            bool: indicate whether levels 0 and 1 of BPM PosX PRBS
                excitation are in desired values.

        """
        level0, level1 = self._handle_bpm_levels(level0, level1)
        for ctl in self._ctlrs.values():
            if not _np.all(ctl.prbs_bpmposx_lvl0 == level0):
                return False
            if not _np.all(ctl.prbs_bpmposx_lvl1 == level1):
                return False
        return True

    def wait_prbs_bpmposx_levels(
            self, level0, level1=None, timeout=_FamFOFBAcqBase.DEF_TIMEOUT):
        """
        Wait for BPM PosX PRBS excitation levels to be equal to desired
        values. Refer to check_prbs_bpmposx_levels for more details on
        arguments and returns.
        """
        _t0 = _time.time()
        while _time.time() - _t0 < timeout:
            if self.check_prbs_bpmposx_levels(level0, level1):
                return True
            _time.sleep(0.1)
        return False

    def set_prbs_bpmposy_levels(self, level0, level1=None):
        """Set BPM PosY PRBS excitation levels.

        Args:
            level0 (numpy.ndarray, (NR_BPM,)):
                desired values for BPM PosX PRBS excitation level 0.
            level1 (optional, numpy.ndarray, (NR_BPM,)):
                If None, we consider level1 = -level0. Defaults to None.

        """
        level0, level1 = self._handle_bpm_levels(level0, level1)
        for ctl in self._ctlrs.values():
            ctl.prbs_bpmposy_lvl0 = level0
            ctl.prbs_bpmposy_lvl1 = level1
        return True

    def check_prbs_bpmposy_levels(self, level0, level1=None):
        """
        Check whether BPM PosY PRBS excitation levels are equal to
        desired values.

        Args:
            level0 (numpy.ndarray, (NR_BPM,)):
                desired values for BPM PosY PRBS excitation level 0.
            level1 (optional, numpy.ndarray, (NR_BPM,)):
                If None, we consider level1 = -level0. Defaults to None.

        Returns:
            bool: indicate whether levels 0 and 1 of BPM PosY PRBS
                excitation are in desired values.

        """
        level0, level1 = self._handle_bpm_levels(level0, level1)
        for ctl in self._ctlrs.values():
            if not _np.all(ctl.prbs_bpmposy_lvl0 == level0):
                return False
            if not _np.all(ctl.prbs_bpmposy_lvl1 == level1):
                return False
        return True

    def wait_prbs_bpmposy_levels(
            self, level0, level1=None, timeout=_FamFOFBAcqBase.DEF_TIMEOUT):
        """
        Wait for BPM PosY PRBS excitation levels to be equal to desired
        values. Refer to check_prbs_bpmposx_levels for more details on
        arguments and returns.
        """
        _t0 = _time.time()
        while _time.time() - _t0 < timeout:
            if self.check_prbs_bpmposy_levels(level0, level1):
                return True
            _time.sleep(0.1)
        return False

    def _handle_bpm_levels(self, level0, level1):
        if len(level0) != NR_BPM or \
                (level1 is not None and len(level1) != NR_BPM):
            raise ValueError('Levels must have length equal to NR_BPM.')
        level0 = _np.asarray(level0)
        if level1 is None:
            level1 = -1 * level0
        # handle FOFB BPM ordering
        level0 = _np.roll(level0, 1)
        level1 = _np.roll(level1, 1)
        return level0, level1

    def cmd_prbs_bpms_enable(self, timeout=_FamFOFBAcqBase.DEF_TIMEOUT):
        """Command to enable PRBS for BPMs."""
        ctrls = list(self._ctlrs.values())
        self._set_devices_propty(ctrls, 'PRBSBPMPosEn-Sel', 1)
        if not self._wait_devices_propty(
                ctrls, 'PRBSBPMPosEn-Sts', 1, timeout=timeout):
            return False
        return True

    def cmd_prbs_bpms_disable(self, timeout=_FamFOFBAcqBase.DEF_TIMEOUT):
        """Command to disable PRBS for BPMs."""
        ctrls = list(self._ctlrs.values())
        self._set_devices_propty(ctrls, 'PRBSBPMPosEn-Sel', 0)
        if not self._wait_devices_propty(
                ctrls, 'PRBSBPMPosEn-Sts', 0, timeout=timeout/2):
            return False
        return True

    def get_data(self, bpmenbl=None, correnbl=None):
        """Get data matrices.

        Returns:
            orbx (numpy.ndarray, NxNR_BPM): Horizontal Orbit.
            orby (numpy.ndarray, NxNR_BPM): Vertical Orbit.
            currdata (numpy.ndarray, NxNR_CORR): Corrector current data.
            kickdata (numpy.ndarray, NxNR_CORR): Corrector kick data.

        """
        # bpm data
        orbx, orby = [], []
        if bpmenbl is None:
            bpmenbl = _np.ones(160, dtype=bool)
        bpmenbl = _np.roll(bpmenbl, 1)
        bpmenbl = bpmenbl.reshape(-1, 8)

        minbpm, mtx = None, None

        for i, ctrl in enumerate(self._ctlrs.values()):
            for idx in range(8):
                if not bpmenbl[i, idx]:
                    continue
                mtx = ctrl.get_bpm_posx(idx)
                mty = ctrl.get_bpm_posy(idx)
                orbx.append(mtx)
                orby.append(mty)

            if mtx is None:
                continue
            if minbpm is None:
                minbpm = mtx.size
            minbpm = _np.min([minbpm, mtx.size, mty.size])

        # # create numpy matrices
        for i, (obx, oby) in enumerate(zip(orbx, orby)):
            orbx[i] = obx[:minbpm]
            orby[i] = oby[:minbpm]
        orbx = _np.array(orbx).T
        orby = _np.array(orby).T

        # # handle FOFB BPM ordering
        orbx = _np.roll(orbx, -1, axis=1)
        orby = _np.roll(orby, -1, axis=1)

        # corrs data
        currdata, kickdata = [], []
        if correnbl is None:
            correnbl = _np.ones(160, dtype=bool)

        mincorr, mtc = None, None

        for i, psn in enumerate(self._psnames):
            if not correnbl[i]:
                continue
            dev = self._psdevs[psn]
            devconv = self._psconv[psn]
            mtc = dev.fofbacc_data
            currdata.append(mtc)
            kickdata.append(devconv.conv_current_2_strength(mtc))

            if mincorr is None:
                mincorr = mtc.size
            mincorr = min(mincorr, mtc.size)

        # # create numpy matrices
        for i, cur in enumerate(currdata):
            currdata[i] = cur[:mincorr]
            kickdata[i] = kickdata[i][:mincorr]
        currdata = _np.array(currdata).T
        kickdata = _np.array(kickdata).T

        return orbx, orby, currdata, kickdata

    def config_acquisition(
            self, nr_points_before: int, nr_points_after=0,
            channel=_FOFBCtrlAcqConst.Channel.sysid_applied,
            repeat=False, external=True) -> int:
        """Configure acquisition for FOFB controllers.

        Args:
            nr_points_before (int): number of points after trigger.
            nr_points_after (int): number of points after trigger.
                Defaults to 0.
            channel (int): acquisition channel (2: sysid, 3: sysid_applied).
                Defaults to 3.
            repeat (bool, optional): Whether or not acquisition should be
                repetitive. Defaults to True.
            external (bool, optional): Whether or not external trigger should
                be used. Defaults to True.

        Returns:
            int: code describing what happened:
                =0: FOFB controllers are ready.
                <0: Index of the first controller which did not stop
                    last acq. plus 1.
                >0: Index of the first controller which is not ready
                    for acq. plus 1.

        """
        if repeat:
            repeat = _FOFBCtrlAcqConst.Repeat.repetitive
        else:
            repeat = _FOFBCtrlAcqConst.Repeat.normal

        if external:
            trig = _FOFBCtrlAcqConst.TrigTyp.external
        else:
            trig = _FOFBCtrlAcqConst.TrigTyp.now

        ret = self.cmd_stop()
        if ret > 0:
            return -ret

        for ctl in self._ctlrs.values():
            ctl.repeat = repeat
            ctl.channel = channel
            ctl.trigger = trig
            ctl.nrsamples_pre = nr_points_before
            ctl.nrsamples_post = nr_points_after

        return self.cmd_start()

    def cmd_stop(self, wait=True, timeout=10) -> int:
        """Stop acquistion.

        Args:
            wait (bool, optional): whether or not to wait controllers
                get ready. Defaults to True.
            timeout (int, optional): Time to wait. Defaults to 10.

        Returns:
            int: code describing what happened:
                =0: FOFB controllers are ready.
                >0: Index of the first FOFB controller which did not
                    update plus 1.

        """
        for ctl in self._ctlrs.values():
            ctl.cmd_ctrl(_FOFBCtrlAcqConst.TrigEvt.stop)

        if wait:
            _time.sleep(2)
            return 0
            # return self.wait_acquisition_finish(timeout=timeout)
        return 0

    def wait_acquisition_finish(self, timeout=10) -> int:
        """Wait for all FOFB controllers to be ready for acquisition.

        Args:
            timeout (int, optional): Time to wait. Defaults to 10.

        Returns:
            int: code describing what happened:
                =0: FOFB controllers are ready.
                >0: Index of the first FOFB controller which did not
                    update plus 1.

        """
        for i, ctl in enumerate(self.ctrldevs.values()):
            t0_ = _time.time()
            if not ctl.wait_acq_finish(timeout):
                return i + 1
            timeout -= _time.time() - t0_
        return 0

    def cmd_start(self, wait=True, timeout=10) -> int:
        """Start acquisition.

        Args:
            wait (bool, optional): whether or not to wait controllers
                get ready. Defaults to True.
            timeout (int, optional): Time to wait. Defaults to 10.

        Returns:
            int: code describing what happened:
                =0: FOFB controllers are ready.
                >0: Index of the first FOFB controller which did not
                    update plus 1.

        """
        for ctl in self._ctlrs.values():
            ctl.cmd_ctrl(_FOFBCtrlAcqConst.TrigEvt.start)

        if wait:
            _time.sleep(2)
            return 0
            # return self.wait_acquisition_start(timeout=timeout)
        return 0

    def wait_acquisition_start(self, timeout=10) -> bool:
        """Wait for all FOFB controllers to be ready for acquisition.

        Args:
            timeout (int, optional): Time to wait. Defaults to 10.

        Returns:
            int: code describing what happened:
                =0: FOFB controllers are ready.
                >0: Index of the first FOFB controller which did not
                    update plus 1.

        """
        for i, bpm in enumerate(self.ctrldevs.values()):
            t0_ = _time.time()
            if not bpm.wait_acq_start(timeout):
                return i + 1
            timeout -= _time.time() - t0_
        return 0

    def update_initial_timestamps(self, bpmenbl=None, correnbl=None):
        """Call this method before acquisition to get data for comparison."""
        self._initial_timestamps = self._get_data_timestamps(bpmenbl, correnbl)

    def wait_update_data(self, timeout=10, bpmenbl=None, correnbl=None) -> int:
        """Call this method after acquisition to check if data was updated.

        For this method to work it is necessary to call
            update_initial_timestamps
        before the acquisition starts, so that a reference for comparison is
        created.

        Args:
            timeout (int, optional): Waiting timeout. Defaults to 10.

        Returns:
            int: code describing what happened:
                -4: initial timestamps were not defined;
                -3: TimeFrame data did not update;
                -2: PRBS data did not update;
                -1: FOFB Acc data did not update;
                =0: data updated.
                >0: index of the first BPM which did not update data plus 1.

        """
        if self._initial_timestamps is None:
            return -4

        tsmp0 = self._initial_timestamps
        while timeout > 0:
            t00 = _time.time()
            tsmp = self._get_data_timestamps(bpmenbl, correnbl)
            errs = [_np.equal(t, t0) for t, t0 in zip(tsmp, tsmp0)]
            continue_ = _np.any(errs[:2]) or _np.any(errs[2]) \
                or _np.any(errs[3:])
            if not continue_:
                return 0
            _time.sleep(0.1)
            timeout -= _time.time() - t00

        # check bpm data timestamps
        errs_bpm = _np.any(errs[:2], axis=0)
        if errs_bpm.any():
            return int(errs_bpm.nonzero()[0][0])+1

        # check timeframe data timestamps
        errs_timeframe = _np.any(errs[-2])
        if errs_timeframe.any():
            return -3

        # check prbs data timestamps
        errs_prbs = _np.any(errs[-1])
        if errs_prbs.any():
            return -2

        # if we reached this point, FOFB Acc data did not update
        return -1

    def _get_data_timestamps(self, bpmenbl=None, correnbl=None):
        """Get data timestamps.

        Returns:
            tsmpx (numpy.ndarray, NxNr BPMs): BPM PosX PVs timestamps.
            tsmpy (numpy.ndarray, NxNr BPMs): BPM PosY PVs timestamps.
            tsmpf (numpy.ndarray, NxNr Corr.): FOFB Acc Data PVs timestamps.
            tsmpt (numpy.ndarray, Nx20): Time Frame Data PVs timestamps.
            tsmpp (numpy.ndarray, Nx20): PRBS Data PVs timestamps.

        """
        # bpm
        tsmpx, tsmpy = [], []
        if bpmenbl is None:
            bpmenbl = _np.ones(160, dtype=bool)
        bpmenbl = _np.roll(bpmenbl, 1)
        bpmenbl = bpmenbl.reshape(-1, 8)

        for i, ctrl in enumerate(self._ctlrs.values()):
            for idx in range(8):
                if not bpmenbl[i, idx]:
                    continue
                pvox = ctrl.pv_object(f'BPM{idx}PosXData')
                pvoy = ctrl.pv_object(f'BPM{idx}PosYData')
                varx = pvox.get_timevars(timeout=self.DEF_TIMEOUT)
                vary = pvoy.get_timevars()
                tsmpx.append(
                    pvox.timestamp if varx is None else varx['timestamp'])
                tsmpy.append(
                    pvoy.timestamp if vary is None else vary['timestamp'])

        # corrs
        tsmpf = []
        if correnbl is None:
            correnbl = _np.ones(160, dtype=bool)

        for i, psn in enumerate(self._psnames):
            if not correnbl[i]:
                continue
            dev = self._psdevs[psn]
            pvof = dev.pv_object('SYSIDFOFBAccData')
            varf = pvof.get_timevars()
            tsmpf.append(pvof.timestamp if varf is None else varf['timestamp'])

        # time frame and prbs
        tsmpt, tsmpp = [], []
        for ctl in self._ctlrs.values():
            pvot = ctl.pv_object('TimeFrameData')
            vart = pvot.get_timevars()
            tsmpt.append(pvot.timestamp if vart is None else vart['timestamp'])
            pvop = ctl.pv_object('PRBSData')
            varp = pvop.get_timevars()
            tsmpp.append(pvop.timestamp if varp is None else varp['timestamp'])

        return tsmpx, tsmpy, tsmpf, tsmpt, tsmpp

    def set_currloop_kp(self, values):
        """Set current loop Kp parameter for all correctors.

        Args:
            values (numpy.ndarray, (160,)):
                Array with values to be applied to correctors Kp
                parameter in psnames order.

        """
        for i, psn in enumerate(self._psnames):
            dev = self._psdevs[psn]
            dev.currloop_kp = values[i]
        return True

    def set_currloop_ti(self, values):
        """Set current loop Ti parameter for all correctors.

        Args:
            values (numpy.ndarray, (160,)):
                Array with values to be applied to correctors Ti
                parameter in psnames order.

        """
        for i, psn in enumerate(self._psnames):
            dev = self._psdevs[psn]
            dev.currloop_ti = values[i]
        return True


# ------- rtmlamp acquisition devices -------

class FOFBCtrlLamp(_FOFBCtrlAcqBase):
    """FOFB controller RTMLAMP acquisition device."""

    _properties = _FOFBCtrlAcqBase._properties

    def __init__(self, devname, auto_monitor_mon=True):
        """Init."""
        # call base class constructor
        super().__init__(
            devname, propty_prefix='LAMP',
            properties=FOFBCtrlLamp._properties,
            auto_monitor_mon=auto_monitor_mon)


class FOFBPSLamp(_Device):
    """FOFB power supply RTMLAMP acquisition device."""

    _properties = (
        'LAMPCurrentRawData', 'LAMPCurrentData',
        'LAMPVoltageRawData', 'LAMPVoltageData',
    )

    def __init__(self, devname, auto_monitor_mon=True):
        """Init."""
        # call base class constructor
        super().__init__(
            devname, properties=FOFBPSLamp._properties,
            auto_monitor_mon=auto_monitor_mon)


class FamFOFBLamp(_FamFOFBAcqBase):
    """FOFB RTMLAMP acquisition device."""

    FOFBCTRL_CLASS = FOFBCtrlLamp
    FOFBPS_CLASS = FOFBPSLamp



"""Simulation."""

import re as _re


class Simulation:
    """Simulation class.

        This simulation class manages simulated epics control process
    variables by registered simulator objects. As a simulator is registered,
    its 'callback_pv_dbase' is executed in order to retrieve the simulator
    dictionary of pvname regular and epics database objects.

        When a SimPV is registered in the simulation, the existing set of
    regular expressions is searched for matches. Simulators that registered
    mathcing regexps are flaged so that callback methods are executed when that
    SimPV is registered or when its state changes (register/get/put actions).
    """

    # TODO: make class thread-safe!

    # pvnames can math only one regexp/simulator?
    PV_DATABASE_UNIQUE = True

    # stores register on/off state
    _register_state = True

    _REGEXP = list()  # pvname regular expression
    _SIMULS = list()  # registered simulators
    _DBASES = list()  # associated epics databases
    _SIMPVS = dict()  # registered SimPVs

    # --- registration methods ---

    @staticmethod
    def simulator_register(simulator):
        """Register simulator(s) in simulation."""
        if not isinstance(simulator, (tuple, list, set)):
            simulator = (simulator, )
        for sim in simulator:
            pvname_regexp = sim.callback_pv_dbase()
            for regexp, dbase in pvname_regexp.items():
                Simulation._pvnames_register(sim, regexp, dbase)

    @staticmethod
    def simulator_unregister(simulator):
        """Unregister simulator(s)."""
        if not isinstance(simulator, (tuple, list, set)):
            simulator = (simulator, )
        regexp, sims, dbases = list(), list(), list()
        for rege, sim, dbas in zip(
                Simulation._REGEXP,
                Simulation._SIMULS,
                Simulation._DBASES):
            if sim not in simulator:
                regexp.append(rege)
                sims.append(sim)
                dbases.append(dbas)
        Simulation._REGEXP, Simulation._SIMULS, Simulation._DBASES = \
            regexp, sims, dbases

    @staticmethod
    def register_state_get():
        """Return registration state."""
        return Simulation._register_state

    @staticmethod
    def register_state_set_on():
        """Set simulation to accept registration of SimPVs."""
        Simulation._register_state = True

    @staticmethod
    def register_state_set_off():
        """Set simulation not to accept registration of SimPVs."""
        Simulation._register_state = False

    # --- PV methods (used mainly in SimPV methods) ---

    @staticmethod
    def pv_register(pvobj):
        """Register SimPV object.

        This is used by SimPV when a sim PV is created.
        """
        pvname = pvobj.pvname
        if pvname in Simulation._SIMPVS:
            # return False if SimPV already registered.
            return False
        sims = set(Simulation.simulator_find(pvname))
        Simulation._SIMPVS[pvname] = (pvobj, sims)
        # execute simulators callback
        for sim in sims:
            sim.callback_pv_register(pvobj)
        return True

    @staticmethod
    def pv_get(pvname, **kwargs):
        """Execute simulators callback functions, prior to SimPV readout.

        Used in SimPV getter methods.
        """
        _, sims, *_ = Simulation._SIMPVS[pvname]
        for sim in sims:
            sim.callback_pv_get(pvname, **kwargs)

    @staticmethod
    def pv_put(pvname, value, **kwargs):
        """Execute simulators callback functions, prior to SimPV setpoint.

        Used in SimPV setter methods. Simulator callback functions must
        return True if setpoint is acceptable in all simulators,
        False otherwise.
        """
        _, sims, *_ = Simulation._SIMPVS[pvname]
        state = True
        for sim in sims:
            state &= sim.callback_pv_put(pvname, value, **kwargs)
        return state

    @staticmethod
    def pv_find(pvname):
        """Return SimPV object.

        Used in SimPV to decide if a new SimPV is needed.
        """
        pv_item = Simulation._SIMPVS.get(pvname, None)
        if not pv_item:
            return None
        pvobj, *_ = pv_item
        return pvobj

    @staticmethod
    def pv_dbase_find(pvname, unique=False):
        """Return databases for a given psname.

        Used in SimPV with unique=True to get the first pvname matching
        database.
        """
        dbases = Simulation._find(pvname, Simulation._DBASES, unique=False)
        if Simulation.PV_DATABASE_UNIQUE and len(dbases) > 1:
            raise ValueError(
                'Database for PV "{}" is not unique!'.format(pvname))
        if not dbases:
            return None
        if unique:
            return dbases.pop()
        return dbases

    # --- utility methods (used mainly in Simulator implementations) ---

    @staticmethod
    def get_pvnames(simulator=None):
        """Return pvnames of registered SimPVs of a given simulator."""
        if simulator is None:
            return set(Simulation._SIMPVS.keys())
        pvnames = set()
        for pvobj, sims in Simulation._SIMPVS:
            for sim in sims:
                if sim == simulator:
                    pvnames.add(pvobj.pvname)
                    break
        return pvnames

    @staticmethod
    def reset():
        """Reset simulation."""
        for sim in Simulation._SIMULS:
            sim.reset()
        Simulation._init()

    @staticmethod
    def simulator_find(pvname, unique=False):
        """Return simulators set for a given pvname."""
        set_ = set()
        for sim in Simulation._SIMULS:
            if sim.pv_check(pvname):
                set_.add(sim)
        return set_.pop() if unique and set_ else set_

    @staticmethod
    def simulator_update(pvname, **kwargs):
        """Execute callback to update/synchronize simulator for pvname."""
        _, sims, *_ = Simulation._SIMPVS[pvname]
        for sim in sims:
            sim.update(pvname, **kwargs)

    @staticmethod
    def pv_check(pvname):
        """Return True of SimPV is registered in simulation."""
        return \
            Simulation._register_state and (
                Simulation.simulator_find(pvname, True) or
                Simulation.pv_find(pvname))

    # --- private methods ---

    @staticmethod
    def _init():
        """Init Simulation class structures."""
        Simulation._REGEXP = list()
        Simulation._SIMULS = list()
        Simulation._DBASES = list()
        Simulation._SIMPVS = dict()

    @staticmethod
    def _find(pvname, itemlist, unique):
        list_ = list()
        for rege, item in zip(Simulation._REGEXP, itemlist):
            if rege.match(pvname):
                list_.append(item)

        # if unique and more than one item, raise exception.
        if unique and len(list_) > 1:
            raise ValueError(
                'Conflicting items for pvname "{}"'.format(pvname))

        # return list
        return list_.pop() if unique and list_ else list_

    @staticmethod
    def _pvnames_register(simulator, pvnames_regexp, dbase):
        """Register simulator pvnames regular expressions."""
        Simulation._REGEXP.append(_re.compile(pvnames_regexp))
        Simulation._SIMULS.append(simulator)
        Simulation._DBASES.append(dbase)



"""Subpackage BSMP.

Basic Small Message protocol

Documentation:
https://github.com/lnls-sirius/libbsmp/blob/master/doc/protocol_v2-30_pt_BR.pdf
https://github.com/lnls-sirius/libbsmp/blob/master/doc/protocol_v2-30_en_US.pdf
"""

from .commands import *
from .entities import *
from .exceptions import *
from .serial import *
from .types import *

__version__ = '2.30.0'
del serial, types, entities, commands, exceptions
